;;; -*- lisp -*-
;;;; $Id: docbook.lisp,v 1.1 2006/07/09 16:16:30 willijar Exp $
;;;; Copyright (C) 2003 Dr. John A.R. Williams, J.A.R.Williams@blueyonder.co.uk

;;;; This file is part of the Common Lisp Markup Package

;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.

;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.

;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
;;;; or retreive a copy from http://www.gnu.org/licenses/gpl.txt

;;;; Process functions for converting s-exp markup to docbook XML

(in-package :markup)

(defun xml-reserved-p (c) (member c '(#\< #\> #\& #\£ #\ñ)))
(defun xml-escape (stream s)
  (loop for c across s
        if (xml-reserved-p c)
        do (format stream "&#~A;" (char-code c))
        else do (write-char c stream)))

(defgeneric docbook(stream tag &optional attr content)
  (:documentation "Convert markup to Docbook XML - returns stream"))

(defmethod docbook(stream tag &optional attr content)
  "Default - just print it"
  (declare (ignore content attr))
  (princ tag stream))

(defmethod docbook(stream (markup list) &optional attr content)
  (declare (ignore content attr))
  (apply #'docbook (cons stream (multiple-value-list (split-markup markup)))))

(defmethod docbook(stream (markup (eql nil)) &optional attr content)
  (declare (ignore attr))
  (dolist (item content) (docbook stream item)))

(defmethod docbook((stream (eql nil)) markup &optional attr content)
  (with-output-to-string(s) (docbook s markup attr content)))

(defmethod docbook(stream (tag string) &optional attr content)
  (declare (ignore content attr))
  (if *escape-html*
      (xml-escape stream tag)
      (write-sequence tag stream)))

(defmethod docbook :around (stream (tag symbol) &optional attr content)
  (let ((*escape-html* (getf attr :escape *escape-html*)))
    (when (block-element-p tag) (write-char #\newline stream))
    (call-next-method stream tag attr content) ))

(defun write-xml-element(stream tagname &optional attr content)
  (write-char #\< stream)
  (write-sequence tagname stream)
  (write-attr attr stream)
  (if (not content)
      (write-sequence "\\>" stream)
      (progn
        (write-char  #\>  stream)
        (dolist (item content) (docbook stream item))
        (write-strings stream "</" tagname ">"))))

(defmethod docbook(stream (tag (eql 'p)) &optional attr content)
  (write-xml-element stream "para" attr content))

(defun collect-dl-entries(content)
  (let ((entries nil)
        (current nil))
    (dolist(item content)
      (let ((tag (split-markup item)))
        (ecase tag
          (dt (when current (push current entries))
               (setf current (cons item nil)))
          (dd (if current
                   (rplacd current item)
                   (setf current (cons nil item)))))))
    (when current (push current entries))
    (nreverse entries)))

(defmethod docbook(stream (tag (eql 'dl)) &optional attr content)
  (declare (ignore attr))
  (write-sequence "<variablelist>
<?dbhtml list-presentation=\"table\"?>
<?dbhtml term-presentation=\"bold\" ?>" stream)
  (dolist(entry (collect-dl-entries content))
    (write-sequence "<varlistentry>" stream)
    (if (car entry)
        (docbook stream (car entry))
        (write-sequence "<term></term>" stream))
    (if (cdr entry)
        (docbook stream (cdr entry))
        (write-sequence "<listitem><para></para></listitem>" stream))
    (write-sequence "</varlistentry>" stream))
  (write-sequence "</variablelist>" stream))

(defmethod docbook(stream (tag (eql 'dt)) &optional attr content)
  (declare (ignore attr))
  (write-xml-element stream "term" nil content))

(defmethod docbook(stream (tag (eql 'dd)) &optional attr content)
  (declare (ignore attr))
  (write-sequence "<listitem>" stream)
  (dolist(item content)
    (if (and (listp item) (block-element-p (markup-tag item)))
        (docbook stream item)
        (write-xml-element stream "para" nil (list item))))
  (write-sequence "</listitem>" stream))

(defmethod docbook(stream (tag (eql 'ul)) &optional attr content)
  (write-xml-element stream "itemizedlist" attr content))

(defmethod docbook(stream (tag (eql 'ol)) &optional attr content)
  (write-xml-element stream "orderedlist" attr content))

(defmethod docbook(stream (tag (eql 'li)) &optional attr content)
  (docbook stream 'dd attr content))

(defmethod docbook(stream (tag (eql 'em)) &optional attr content)
  (write-xml-element stream "emphasis" attr content))

(defmethod docbook(stream (tag (eql 'a)) &optional attr content)
  (write-xml-element stream "ulink" `(:|url| ,(getf attr :href)) content))
