;;;; -*- lisp -*-
;;;; $Id: figures.lisp,v 1.5 2006/08/18 08:48:08 willijar Exp $
;;;; Copyright (C) 2002 Dr. John A.R. Williams, J.A.R.Williams@blueyonder.co.uk

;;;; This file is part of the Common Lisp Markup Package

;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.

;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.

;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
;;;; or retreive a copy from http://www.gnu.org/licenses/gpl.txt

;;;; helper functions for preprocessing markup for web presentation

(in-package :markup)

(defvar *run-external-programs* t
  "If true external programmes will be run to generate/convert content")

(defun run-shell(&rest args)
  (when *run-external-programs*
    (apply #'jarw.port:run-shell args)))

(defun copy-file(src dest)
  (run-shell (format nil "cp ~S ~S" (namestring src) (namestring  dest))))

(defun source-changed-p(src dest)
  "Returns true if destination needs to be updated because src has changed."
  (and (probe-file src)
       (or (not (probe-file dest))
           (> (file-write-date src) (file-write-date dest)))))

(defparameter latexBegin
"\\documentclass[12pt]{article}
\\usepackage{amsmath}
\\pagestyle{empty}
\\begin{document}
\\newcommand{\\mat}[1]{{\\bf #1}}
\\newcommand{\\erf}{\\mbox{erf}}
\\newcommand{\\erfc}{\\mbox{erfc}}
\\newcommand{\\sinc}{\\mbox{sinc}}
"
"Initial component of latex document")

(defparameter latexEnd
"\\end{document}"
"Final part of latex document")

(defvar *image-pixel-width* 700
  "The base width for image sizes in pixels")

;; helper functions used to construct commands for shell

(defun image-out-cmd(dest &key (scale nil) (width 1.0) (margin 1))
  "construct string of commands for shell to create output to given file"
  (when width
    (setq scale nil)
    (setq width (floor (* *image-pixel-width* width))))
  (format nil
          "pnmcrop -white | pnmmargin -white ~D | pnmscale ~:[-width=~D~;~:*~F~*~] | ~A > \"~A\""
          margin
          scale
          width
          (cdr (assoc (pathname-type dest)
                      '(("gif" . "ppmquant 256 | ppmtogif -interlace -")
                        ("png" . "pnmtopng -")
                        ("eps" . "pnmtops -noturn -nosetpage -equalpixels -rle -")
                        ("jpeg" . "pnmtojpeg -")
                        ("jpg" . "pnmtojpeg -")) :test #'string-equal))
          dest))

(defun image-to-pnm-cmd(src &key (dpi 300))
  "construct string of commands for shell to create output to given file"
  (let ((type (pathname-type src)))
    (if (member type '("ps" "eps" "pdf") :test #'string-equal)
        (format nil
                "gs -sDEVICE=ppmraw -sOutputFile=- -r~D -q -dNOPAUSE \"~A\" < /dev/null | "
                dpi src)
        (format nil "~A < \"~A\" | "
                (cdr (assoc type
                            '(("gif" . "giftopnm")
                              ("png" . "pngtopnm")
                              ("jpg" . "jpegtopnm")
                              ("jpeg" . "jpegtopnm")
                              ("ps" . "pstopnm -stdout")
                              ("eps" . "pstopnm -stdout")
                              ("pdf" . "pstopnm -stdout"))
                            :test #'string-equal))
                src))))

(defun ps-to-pnm-cmd(src &key size (dpi 300))
  "Construct string of commands for shell to read a ps file and pipe out
   a ppm file for further processing"
  (format nil "gs -sDEVICE=ppmraw -sOutputFile=- ~@[-g~Dx~:*~D~] -r~D -q -dNOPAUSE \"~A\" < /dev/null | " size dpi src))

(defun latex-to-ps-cmd(src &optional
                       (dest (translate-pathname src "*.*" "*.ps")))
  "String of commands to convert a latex file to a postscript file"
  (format
   nil
   "cd ~S; latex -interaction=nonstopmode \"~A\"; dvips -f \"~A\" -o \"~A\"; "
   (directory-namestring src)
   src
   (translate-pathname src "*.*" "*.dvi")
   dest))

(defun latex-to-pdf-cmd(src &optional
                        (dest (translate-pathname src "*.*" "*.pdf")))
  "String of commands to convert a latex file to a postscript file"
  (format
   nil
   "cd ~S; latex -interaction=nonstopmode \"~A\"; dvipdf \"~A\" \"~A\"; "
   (directory-namestring src)
   src
   (translate-pathname src "*.*" "*.dvi")
   dest))

(defun latex-to-image-file(code dest &key (dpi 100) (detail 2.5) (margin 1))
  "Given some latex code generate a gif image file."
  (let ((is-ps (or (string-equal (pathname-type dest) "ps")
                   (string-equal (pathname-type dest) "eps"))))
    (with-temp-pathname
        (texpath :prefix "tmp-latex-file-":type "tex"
                 :delete-all-types t)
      (let ((pspath (if is-ps dest (translate-pathname texpath "*.*" "*.ps"))))
        (with-open-file
            (stream texpath :direction :output :if-exists :supersede)
          (write-string (concatenate 'string latexBegin code latexEnd)
                        stream))
        (run-shell (latex-to-ps-cmd texpath pspath))
        (unless is-ps
          (let ((work-dpi (* detail dpi))
                (size (round (* 12 detail dpi)))
                (reduct-ratio (/ 1.0 detail)))
            (run-shell
             (concatenate 'string
                          (ps-to-pnm-cmd pspath :size size :dpi work-dpi)
                          (image-out-cmd dest :scale reduct-ratio :width nil
                                         :margin margin)))))))))

(defun ps-to-image-file(src dest &key (detail 2.5)
                        (mag 1.0) (width 1.0) (dpi 100) (margin 1))
  (let ((work-dpi (* detail dpi))
        (size (round (* 12 detail dpi)))
        (reduct-ratio (/ mag detail)))
    (run-shell
     (concatenate 'string
                  (ps-to-pnm-cmd src :dpi work-dpi :size size)
                  (image-out-cmd dest :scale reduct-ratio :width width
                                 :margin margin)))))

(defun image-to-image-file(src dest &key
                           width (dpi 300) (mag 1.0) (margin 1))
  (run-shell
   (concatenate 'string
                (image-to-pnm-cmd src :dpi dpi)
                (image-out-cmd dest :scale mag :width width
                               :margin margin))))

(defun get-bbox(src)
  (with-output-to-string(os)
    (run-shell "gs -sDEVICE=bbox -dNOPAUSE -dBATCH -q $0"
               :args (list (namestring src))
               :error os) ))

(defun set-bbox(src &optional
                (dest (translate-pathname src "*.ps" "*.eps"))
                (bbox (get-bbox src)))
  (flet ((begins-with(a b)
           (when (> (length a) (length b))
             (string-equal a b :end1 (length b)))))
    (with-open-file (in src :direction :input)
      (with-open-file(out dest :direction :output :if-exists :supersede)
        (let ((line (read-line in nil nil)))
          (if (search "EPSF-3.0" line)
              (write-line line out)
              (write-line "%!PS-Adobe-3.0 EPSF-3.0" out)))
        (write-line bbox out)
        (while
            (let ((line (read-line in nil nil)))
              (cond ((begins-with line "@PJL"))
                    ((begins-with line "%%BoundingBox:"))
                    (t (write-line line out)))
              line))))))

(defun fig-to-eps-file(src &optional
                       (dest (translate-pathname src "*.*" "*.eps"))
                       &key  latex)
  (when (source-changed-p src dest)
    (if latex
        (with-temp-pathname (pstex-t-path :prefix "tmp-latex-file-"
                                          :type "pstex_t"
                                          :delete-all-types t)
          (let ((pstex-path (translate-pathname pstex-t-path "*.*" "*.pstex"))
                (ps-path (translate-pathname pstex-t-path "*.*" "*.ps")))
            (run-shell
             (format nil "fig2dev -L pstex -p \"~A\" \"~A\" \"~A\"; fig2dev -L pstex_t -p \"~A\" \"~A\" \"~A\""
                     pstex-path src pstex-path
                     pstex-path src pstex-t-path) )
            (latex-to-image-file (format nil "\\input{~A}" pstex-t-path)
                                 ps-path)
            (set-bbox ps-path dest)))
        (run-shell (format nil "fig2dev -L eps \"~A\" \"~A\"" src dest)))))

(defun fig-to-image-file(src dest &key (margin 1) (width 1.0) latex)
  (let ((eps-file (translate-pathname src "*.*" "*.eps")))
    (fig-to-eps-file src eps-file :latex latex)
    (ps-to-image-file eps-file dest :width width :margin margin)))


(defun convert-markup-file(src dest)
  (let ((markup (with-open-file(is src) (preprocess-html (read is))))
        (type (pathname-type dest)))
    (cond ((equalp type "html")
           (with-open-file(os dest :direction :output)
             (html os markup)))
          ((equalp type "tex")
           (with-open-file(os dest :direction :output)
             (latex os markup)))
          ((equalp type "pdf")
           (let ((texdest (translate-pathname dest "*.*" "*.tex")))
             (with-open-file (os texdest  :direction :output)
               (latex os markup))
             (run-shell (latex-to-pdf-cmd texdest dest))))
          ((equalp type "ps")
           (let ((texdest (translate-pathname dest "*.*" "*.tex")))
             (with-open-file (os texdest  :direction :output)
               (latex os markup))
             (run-shell (latex-to-ps-cmd texdest dest))))
          (t (error "Unknown file format")))))