;;;; -*- lisp -*-
;;;; $Id: html-preprocess.lisp,v 1.2 2007/07/09 19:24:48 willijar Exp $
;;;; Copyright (C) 2002 Dr. John A.R. Williams, J.A.R.Williams@blueyonder.co.uk

;;;; This file is part of the Common Lisp Markup Package

;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.

;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.

;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
;;;; or retreive a copy from http://www.gnu.org/licenses/gpl.txt

;;;; Preprocessing functionality for markup to html in an environment

(in-package :markup)

(defgeneric preprocess-html(tag &optional attr content)
  (:documentation "Preprocesses markup for html rendering
     - returns modified markup and causes necessary sideffects"))

(defmethod preprocess-html(markup &optional attr content)
  "By default objects are returned as is"
  (declare (ignore attr content))
  markup)

(defmethod preprocess-html :around ((tag symbol) &optional attr content)
  (let* ((markup (call-next-method tag attr content))
	 (attr (if (listp (car markup)) (cdar markup)))
	 (id (getf attr :id)))
    (when id
      (setf (gethash id *references*)
	    (list (getf attr :label (getf attr :id))
		  *document-name* id
		  (multiple-value-bind(tag attrs content)
		      (split-markup markup)
		    (let ((attrs (copy-list attrs)))
		      (remf attrs :id)
		      (join-markup tag attrs content))))))
    markup))

(defmethod preprocess-html((markup list) &optional attr content)
  "Given a list - split into components and output"
  (declare (ignore content attr))
	(when markup
		(apply #'preprocess-html (multiple-value-list (split-markup markup)))))

(defmethod preprocess-html((tag symbol) &optional attr content)
  "Default is output markup as before and preprocess content"
  (join-markup tag attr
	       (mapcar #'(lambda(item)
			   (if (symbolp item)
			       item ; symbols by themselves are left alone
			     (preprocess-html item)))
		       content)))

(defmethod preprocess-html((tag (eql 'section)) &optional attrs content)
  "Might have items whose processing is level sensitive"
  (let ((*section-level* (getf attrs :level (1+ *section-level*))))
    (join-markup tag attrs (mapcar #'preprocess-html content))))

(defun is-latex-math-p(math)
  "returns t if the math text requires latex rendering"
  (when (find-if #'(lambda(char) (member char '(#\\ #\_ #\^ #\{))) math)
    t))

(defun preprocess-media(attrs source &key type (output "image/png") args)
  "Register media and output type conversion, return the modified attrs"
  (when (pathnamep source) (setf source (find-file source)))
  (let ((key
         (if type
             (register-media source *media-server* :content-type type)
             (register-media source *media-server*))))
    (media-variant key *media-server* output args)
    `(:media ,key ,@attrs)))

(defmethod preprocess-html((tag (eql 'math)) &optional attrs content)
  "Content preprocessor for math, will create an gif image of the
content using the latex math environment if the :latex attribute is
true or content contains \\. The latex content is cached so if you
have the same content several times only one image is generated and
referred to several times. Otherwise a span of class math is put around
the contents."
  (let ((math (first content)))
    (join-markup tag
                 (if (is-latex-math-p math)
                     (preprocess-media attrs (strcat "$" math "$")
                                       :type "text/x-eqn")
                     attrs)
                 content)))

(defun is-eqnarray-p(math)
  "returns true if the math text requires eqnarray rather than equation"
  (dolist (reg (regex:scan-str-multiple
		(load-time-value (compile-str "&|\\begin.*?\\end") t) math))
    (let ((reg (elt reg 0)))
      (when (= (- (cdr reg) (car reg)) 1)
	(return-from is-eqnarray-p t)))))

(defmethod preprocess-html((tag (eql 'equation)) &optional attrs content)
  "Content preprocessor for equation, will create an gif image of the
     latex content. Content must set up the required latex environment.
     The image for the latex content is cached so if you have the
     same content several times only one image is generated. Output puts
     an image in a table with a label on the right hand column."
  (let* ((math (first content))
         (mathtype (if (is-eqnarray-p math) "eqnarray*" "displaymath")))
    (join-markup
     tag
     `(:label ,(incf *equation-counter*)
       ,@(preprocess-media attrs
         (if (getf attrs :raw)
             math
             (strcat "\\begin{" mathtype "}"
                     math
                     "\\end{" mathtype "}"))
         :type "text/x-eqn"))
     content)))

(defmethod preprocess-html((tag (eql 'figure)) &optional attrs content)
  (join-markup tag (append (list :label (incf *figure-counter*)) attrs)
	     (mapcar #'preprocess-html content)))

(defmethod preprocess-html((tag (eql 'graphic)) &optional attrs content)
  (join-markup
   tag
   (preprocess-media attrs (pathname (first content)))
   content))

(defmethod preprocess-html((tag (eql 'file)) &optional attrs content)
    (join-markup
     tag
     (preprocess-media attrs (pathname (first content)))
     content))

(defmethod preprocess-html((tag (eql 'structured-text))
			   &optional attrs content)
  "Structured text needs to store context information for its elements"
  (join-markup tag
	       (append (list :figure-counter *figure-counter*
			     :equation-counter *equation-counter*
			     :document-name *document-name*
			     :file-counter *file-counter*)
		       attrs)
	       (mapcar #'preprocess-html content)))

(defmethod preprocess-html((tag (eql 'cite)) &optional attr content)
  "If two values are given this is a bibliography definition and it
   is added to the bibliography of the current tutor"
  (when (rest content)
    (setf (gethash (first content) *bibliography*)
	  (rest content)))
  (join-markup tag attr content))