;;;; -*- lisp -*-
;;;; $Id: html.lisp,v 1.1 2008/05/08 09:33:19 willijar Exp $
;;;; Copyright (C) 2002 Dr. John A.R. Williams, J.A.R.Williams@blueyonder.co.uk

;;;; This file is part of the Common Lisp Markup Package

;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.

;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.

;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
;;;; or retreive a copy from http://www.gnu.org/licenses/gpl.txt

;;;; Process functions for converting s-exp markup to html

(in-package :markup)

;;; dynamic variables controling conversion of markup to html
;;; If need be set these using with-markup-environment
;;; many are also used during preprocessing
(defvar *section-level* 0 "Section level")
(defvar *figure-counter* 0 "Number of last figure")
(defvar *equation-counter* 0 "Number of last equation")
(defvar *file-counter* 0 "Number of last file object")
(defvar *references* (make-hash-table)
  "Dictionary of labeled content for cross references")
(defvar *bibliography* (make-hash-table)
  "Dictionary of bibliographic items")
(defvar *escape-html* t
  "If true escape special html chars in strings, Set using :escape attribute")
(defvar *document-name* "unnamed" "Name of current document")
(defvar *document-title* nil "Title of current document")
(defvar *document-author* nil "Name of current document")
(defvar *document-date* nil "Date of current document")

(defvar *footnotes* nil "The list of accumulated footenotes")
(defvar *ref-suffix* "" "Suffix to add on to anchor links")
(defvar *anchor-ids* nil
  "If true a named anchor will prefix all elelements with the id attribute - to support very old browsers that don't understand the id attribute")

(defmacro with-new-document(name &body body)
  "Established counters for a new html page"
  `(let ((*figure-counter* 0)
         (*equation-counter* 0)
         (*file-counter* 0)
         (*section-level* 0)
         (*footnotes* nil)
         (*document-name* ,name))
    ,@body))

(declaim (inline filename))
(defun filename(name count type &key (label "fig"))
  "Filename component for auxilliary file determined by document name,
file number and type"
  (format nil "~A-~A-~3,'0D.~A" (string-downcase name) label count type))

(defun html-reserved-p (c) (member c '(#\< #\> #\&)))
(defun empty-element-p (tag) (member tag '(meta img br hr input)))
(defun inline-element-p(tag) (member tag '(math a img input strong b em footnote)))
(defun block-element-p(tag) (not (inline-element-p tag)))

(defun html-escape (stream s)
  (loop for c across s
        if (html-reserved-p c)
        do (format stream "&#~A;" (char-code c))
        else do (write-char c stream)))

(defun is-boolean-attribute(attr)
  (member attr '(:checked :disabled :readonly :ismap)))
(defun is-ignored-attribute(attr)
  (member attr '(:escape :level)))

(defgeneric format-attr(stream attr value)
  (:documentation "Specialisation on formatting an attribute")
  (:method (stream attr value)
    (unless (is-ignored-attribute attr)
      (write-strings stream " " (symbol-name attr))
      (unless (is-boolean-attribute attr)
        (write-string "=\"" stream)
        (if (stringp value)
            (loop for c across value
                  if (or (html-reserved-p c) (eql c #\"))
                  do (format stream "&#~A;" (char-code c))
                  else do (write-char c stream))
            (write-strings stream value))
        (write-char #\" stream)))))

(defmethod format-attr(stream (attr (eql :width)) (value number))
  (let ((fractional (< value 1)))
    (format stream " ~A=\"~:[~D~;~D%~]\"" (symbol-name attr)
            (< value 1)
            (if fractional (floor (* 100 value)) value))))

(defun write-attr (attr stream)
  (declare (optimize (speed 3)))
  (loop for a on attr by #'cddr
        do (case (car a)
             (:value (format-attr stream (car a)
                                  (format-output (getf attr :datatype) (cadr a))))
             (:datatype)
             (t (format-attr stream (car a) (cadr a))))))

(defun write-strings(stream &rest args)
  "Stringize arguments onto stream"
  (declare (optimize speed) (stream stream))
  (dolist (item args)
    (typecase item
      (string (write-sequence item stream))
      (list (apply #'write-strings (cons stream item)))
      (t (princ item stream))))
  args)

(defgeneric html(stream tag &optional attr content)
  (:documentation "Convert markup to html - returns stream"))

#|
(defmethod html :around (stream tag &optional attr content)
  "Always return stream"
  (declare (ignore tag attr content))
  (call-next-method) stream)
|#

(defmethod html(stream tag &optional attr content)
  "Default - just print it"
  (declare (ignore content attr))
  (princ tag stream))

(defmethod html(stream (markup list) &optional attr content)
  (declare (ignore content attr))
  (apply #'html (cons stream (multiple-value-list (split-markup markup)))))

(defmethod html(stream (markup (eql nil)) &optional attr content)
  (declare (ignore attr))
  (dolist (item content) (html stream item)))

(defmethod html((stream (eql nil)) markup &optional attr content)
  (with-output-to-string(s) (html s markup attr content)))

(defmethod html(stream (tag string) &optional attr content)
  (declare (ignore content attr))
  (if *escape-html*
      (html-escape stream tag)
      (write-sequence tag stream)))

(defmethod html((path pathname) markup  &optional attr content)
  (with-open-file(os path :direction :output :if-exists :supersede)
    (call-next-method os markup attr content)))

(defmethod html :around (stream (tag symbol) &optional attr content)
  (let ((*escape-html* (getf attr :escape *escape-html*))
        (id (getf attr :id)))
    (when (block-element-p tag) (write-char #\newline stream))
    (when (and id *anchor-ids*) (format stream "<a name=\"~A\"></a>" id))
    (call-next-method stream tag attr content) ))

(defmethod html(stream (tag symbol) &optional attr content)
  (let ((tagname (symbol-name tag)))
    (write-char #\< stream)
    (write-sequence tagname stream)
    (write-attr attr stream)
    (write-char  #\>  stream)
    (unless (empty-element-p tag)
      (dolist (item content) (html stream item))
      (write-strings stream "</" tagname ">"))))

(defmethod html(stream (tag (eql 'html)) &optional attr content)
  (declare (ignore content attr))
  (write-line
   "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">" stream)
  (call-next-method))

(defmethod html(stream (tag (eql 'document)) &optional attr content)
	(html stream 'html attr content))

(defmethod html(stream (tag (eql 'body)) &optional attrs content)
	(declare (ignore attrs content))
	(let ((*section-level* 0))
	  (call-next-method)))

(defmethod html(stream (tag (eql 'title)) &optional attr content)
	(setq *document-title* (first content))
	(call-next-method stream tag attr content))

(defmethod html(stream (tag (eql 'author)) &optional attr content)
	(declare (ignore attr))
	(setq *document-author* (first content))
	(html stream `((meta :name "author" :contents ,(first content)))))

(defmethod html(stream (tag (eql 'date)) &optional attr content)
	(declare (ignore attr))
	(setq *document-date* (first content))
	(html stream `((meta :name "date" :contents ,(first content)))))

(declaim (inline value-string))
(defun value-string(attrs)
  (format *debug-io* "datatype=~A value=~A output=~S"
         (getf attrs :datatype)  (getf attrs :value)
         (format-output (getf attrs :datatype) (getf attrs :value)))
  (format-output (getf attrs :datatype) (getf attrs :value)))

(defmethod html(stream (tag (eql 'input)) &optional attrs content)
  (declare (ignore content))
  (if (eq (getf attrs :disabled) :text)
      (unless (equalp (string (getf attrs :type)) "submit")
        (html stream
              (format-output (getf attrs :datatype) (getf attrs :value))))
      (call-next-method)))

(defmethod html(stream (tag (eql 'maq)) &optional attrs content)
  (let ((value (getf attrs :value)))
    (let ((value  (if (listp value) value (list value)))
          (name (getf attrs :name))
          (class '(:class "maq"))
          (content (mapcar
                    #'(lambda(item)
                        (if (consp item) item
                            (cons item (format nil "~A" item))))
                    content)))
      (if (eq (getf attrs :disabled) :text)
          (html stream (join-strings
                        (mapcar
                         #'(lambda(value)
                             (or (cdr (assoc value content :test #'equal))
                                 value))
                         value)
                        ", "))
          (flet ((check (choice)
                   `((input  :type "check" ,@(when (getf attrs :disabled)
                                                   (list :disabled t))
                      :name ,name :value ,(or (car choice) "NIL")
                      ,@(when (member (car choice) value :test #'equal)
                              '(:checked 1) )))))
            (case (getf attrs :style)
              (:horizontal
               (let ((twidth (getf attrs :width "80%"))
                     (cwidth (format nil "~A%" (/ 100.0 (length content)))))
                 (html stream 'table (append (list :width twidth) class)
                       `((tr
                          ,@(mapcar
                             #'(lambda(choice)
                                 `((td :width ,cwidth :align "center" :valign "top")
                                   ,(cdr choice)))
                             content))
                         (tr
                          ,@(mapcar #'(lambda(choice)
                                        `((td :align "center") ,(check choice)))
                                    content))))))
              (:dropdown
               (html stream 'select
                     (append  `(:multiple "multiple"
                                :name ,name
                                ,@(when (getf attrs :disabled)
                                        (list :disabled t))
                                ,@(when (getf attrs :size)
                                        (list :size (getf attrs :size)))) class)
                     (mapcar
                      #'(lambda(choice)
                          `((option :value ,(car choice)
                             ,@(when (member (car choice) value :test #'equal)
                                     '(:selected "selected")))
                            ,(cdr choice)))
                      content)))
              (otherwise
               (html
                stream 'div '(:class "maq")
                (mapcan
                 #'(lambda(choice)
                     (let ((v (if (consp choice) (car choice) choice)))
                       (list '(br)
                             `((input :name  ,name
                                :type "checkbox"
                                :value ,v
                                ,@(when (getf attrs :disabled) '(:disabled t))
                                ,@(when (and value (find v value :test #'equal))
                                        '(:checked t) )))
                             (if (consp choice) (cdr choice) choice) )))
                 content )))))))))

(defmethod html(stream (tag (eql 'mcq)) &optional attrs content)
  "Define a multiple choice question tag"
  (let ((value (getf attrs :value))
        (name (getf attrs :name))
        (class '(:class "mcq"))
        (content (mapcar
                  #'(lambda(item)
                      (if (consp item) item
                          (cons item (format nil "~A" item))))
                  content)))
    (if (eq (getf attrs :disabled) :text)
        (html stream  (cdr (assoc value content :test #'equal)))
        (flet ((radio (choice)
                 `((input :type "radio"
                     ,@(when (equal (car choice) value)
                             '(:checked "checked") )
                    ,@(when (getf attrs :disabled) (list :disabled t))
                    ,@(when (getf attrs :readonly) (list :readonly t))
                    :name ,name :value ,(or (car choice) "NIL")))))
          (case (getf attrs :style)
            (:horizontal
             (let ((twidth (getf attrs :width "80%"))
                   (cwidth (format nil "~A%" (/ 100.0 (length content)))))
               (html stream 'table (append (list :class "mcq" :width twidth
                                                 :cellpadding "3")
                                           class)
                     `((tr
                        ,@(mapcar
                           #'(lambda(choice)
                               `((th :width ,cwidth :align "center"
                                  :valign "top")
                                 ,(cdr choice)))
                           content))
                       (tr
                        ,@(mapcar #'(lambda(choice)
                                      `((td :align "center") ,(radio choice)))
                                  content))))))
            (:dropdown
             (html stream 'select
                   (append  `(:name ,name
                              ,@(when (getf attrs :disabled) (list :disabled t))
                              ,@(when (getf attrs :size)
                                      (list :size (getf attrs :size)))) class)
                   (if (getf attrs :readonly)
                       `(((option :value ,value :selected "selected")
                          ,(cdr (first (member value content
                                               :key #'car :test #'equal)))))
                       (mapcar
                        #'(lambda(choice)
                            `((option :value ,(car choice)
                               ,@(when (equal (car choice) value)
                                       '(:selected "selected")))
                              ,(cdr choice)))
                        content))))
            (otherwise
             (html stream 'div class
                   (mapcan #'(lambda(choice)
                               (list '(br) (radio choice) (cdr choice)))
                           content ))))))))

(defmethod html(stream (tag (eql 'boolean))  &optional attrs content)
  (declare (ignore content))
  (let ((value (getf attrs :value)))
    (when (and (stringp value) (= 0 (length value)))
      (setf value nil))
    (if (getf attrs :disabled)
        (html stream (if value "Yes" "No"))
        (format stream
                "<input name=\"~A\" type=\"checkbox\" ~@[checked=\"checked\"~] />"
                (getf attrs :name) value))))

(defmethod html(stream (tag (eql 'rst)) &optional attrs content)
  (declare (ignore attrs))
  (when (first content)
    (html stream (parse-structured-text (first content)))))

(defmethod html(stream (tag (eql 'textarea)) &optional attrs content)
  (let ((value (getf attrs :value)))
    (when value
      (setf attrs (remove-attribute :value attrs))
      (setf content (list (format-output (getf attrs :datatype) value))))
    (if (getf attrs :disabled)
        (html stream `(section
                       ,(if (getf attrs :structured-text)

                            (when value `(rst ,value))
                            (format-output (getf attrs :datatype) value))))
        (let((txt (car content))) ;; try and determine how many rows are needed
          (when (and (not(getf attrs :rows))  (stringp txt))
            (setf (getf attrs :rows)
                  (+ 2 (length (jarw.string:word-wrap
                                txt
                                :length (getf attrs :cols 60))))))
          (call-next-method stream tag attrs content)))))

(defmethod html(stream (tag (eql 'section)) &optional attrs content)
  (let ((*section-level* (or (getf attrs :level) (1+ *section-level*)))
        (title (getf attrs :title))
        (class (getf attrs :class)))
    (format
     stream
     "~%<DIV ~A CLASS=\"~:[section-~A~;~:*~A~1*~]\">~%~@[<H~:*~A>~A</H~2:*~A>~]"
     (with-output-to-string(out) (write-attr attrs out))
     class *section-level*
     (if (listp title)
         (html stream nil nil title)
         title))
    (dolist (item content) (html stream item))
    (write-line "</DIV>" stream)))

(defun html-media-url(attrs source &key type (output"image/png") args)
  (let* ((media-key
          (or (getf attrs :media)
              (getf (preprocess-media
                    attrs source
                    :type type
                    :output output
                    :args args)
                    :media))))
    (media-variant-url
     (media-variant media-key *media-server* output args)
     *media-server*)))

(defmethod html (stream (tag (eql 'math)) &optional attrs content)
  (let ((math (first content)))
    (if (is-latex-math-p math)
        (format stream "~%<img class=\"math\" src=\"~A\" alt=\"~A\" />~%"
                (html-media-url attrs (strcat "$" math "$")
                                :type "text/x-eqn")
                math)
        (format stream "~%<span class=\"math\">~A</span>~%" math))))

(defmethod html (stream (tag (eql 'equation)) &optional attrs content)
  (let ((math (first content)))
    (unless (getf attrs :raw)
      (setf math
            (let ((mathtype (if (is-eqnarray-p math)
                                "eqnarray*" "displaymath")))
              (strcat "\\begin{" mathtype "}" math "\\end{" mathtype "}"))))
      (format
       stream
       "~%<table class=\"equation\" width=\"100%\"><tr>
<td align=\"center\"><img class=\"equation\" src=\"~A\" alt=\"~A\"></td>
<td align=\"right\">(Eqn. ~D)</td></tr></table>"
       (html-media-url attrs math  :type "text/x-eqn")
       math
       (or (getf attrs :label) (incf *equation-counter*)))))

(defmethod html (stream (tag (eql 'figure)) &optional attrs content)
  (let ((caption (getf attrs :caption))
        (style (format nil "float:~A; clear: ~:*~A;"
                       (getf attrs :align "none")))
        (no (or (getf attrs :label) (incf *figure-counter*))))
    (html stream 'div `(:class "figure" :style ,style)
          (append content  `(((p :class "caption" :style "clear:both;")
                              "Figure " ,no ": "
                              ,@(if (listp caption)
                                    caption
                                    (list caption))))))))

(defmethod html (stream (tag (eql 'graphic)) &optional attrs content)
  (format stream
          "~%<img src=\"~A\" style=\"float:~A; clear:~:*~A;\" alt=\"~A\">"
          (html-media-url attrs (pathname (first content))
                          :output (getf attrs :type "image/png"))
          (getf attrs :align "none") (first content)))

(defmethod html (stream (tag (eql 'file)) &optional attrs content)
  (format stream "~%<a href=\"~A\">~A</a>"
          (html-media-url attrs (pathname (first content)) :output nil)
          (first content)))

(defun navbar(menus &key on-url (relative "") (style :horizontal-table))
  (let* ((max-size (apply #'max (mapcar #'length menus)))
         (cell-size (/ 100 max-size)))
    (ccase style
      (:horizontal-table
       `((table :width "100%" :class "topmenu"
          :cellpadding 2 :cellspacing 1 :border 0)
         ,@(mapcan
            #'(lambda(menu)
                (when menu
                `(((tr :escape nil)
                  ,@(mapcar
                     #'(lambda(item)
                         `((td :align "center" :valign "top"
                            :width ,(format nil "~3f%" cell-size)
                            :class ,(if (equal (first item) on-url)
                                        "navon" "navoff"))
                           ,(if item
                                `((a :href ,(strcat relative
                                                    (first item)))
                                  ,(second item))
                                "&nbsp;")))
                     (append menu
                             (make-sequence
                              'list
                              (- max-size (length menu))
                              :initial-element nil)) )))))
            menus))))))

(defmethod html (stream (tag (eql 'navbar)) &optional attrs content)
  (html stream (apply #'navbar (append content attrs))))

(defmethod html(stream (tag (eql 'ref)) &optional attrs content)
  (let* ((id (first content))
         (ref (gethash id *references*)))
    (when ref
      (if (and (getf attrs :inplace) (fourth ref))
          (html stream (fourth ref))
          (format stream
                  "&nbsp;<a href=\"~A~@[~A~]~@[#~A~]\" title=\"~A\" ~@[target=\"~A\"~]>~A</a> "
                  (second ref)          ; path
                  *ref-suffix*
                  (third ref)           ; internal ref
                  (first ref)           ; real title
                  (getf attrs :target)  ; target if specified
                  (or (second content) (first ref))))))) ; link text
#|
(defun st-inline-picture(things prefix)
  (markup::process-inline-text
   things
   #'(lambda(txt)
       (regex:regex-process-string
	(load-time-value
	 (compile-str"\"([^\"]+)\":picture:([^\\ \\n\\r]+)") t) txt
	 #'(lambda(regs)
	     (list (list 'picture :src (concatenate 'string prefix (elt regs 2))
			 :alt (elt regs 1)))) ))))
|#
(defmethod html(stream (tag (eql 'structured-text)) &optional attrs content)
  "Context is restored as structured text elements won't have attributes from
preprocessing"
  (let ((*figure-counter* (getf attrs :figure-counter *figure-counter*))
        (*equation-counter* (getf attrs :equation-counter *equation-counter*))
        (*document-name* (getf attrs :document-name *document-name*))
        (*file-counter* (getf attrs :file-counter *file-counter*)))
    (dolist (item content)
      (html stream (parse-structured-text item)))))

(defmethod html(stream (tag (eql 'cite)) &optional attr content)
  "Citations produce link back to bibliography"
  (declare (ignore attr))
  (format stream " [<a href=\"./bibliography/~A\">~:*~A</a>] "
          (first content)))

(defmethod html(stream (tag (eql 'td)) &optional attr content)
  "Citations produce link back to bibliography"
  (when (and (car content) (not (cdr content)) (numberp (car content))
           (not (getf attr :align)))
      (setf (getf attr :align) :right))
  (call-next-method stream tag attr content))

(defmethod html(stream (tag (eql 'form)) &optional attr content)
  (setf (getf attr :accept-charset) "UTF-8"
        (getf attr :enctype) "multipart/form-data")
  (call-next-method stream tag attr content))

(defmethod html(stream (tag (eql 'footnote)) &optional attr content)
  (push (first content) *footnotes*)
  (html stream `((a :href "#footnotes" :title ,(first content))
                 ((sup :class :footnote ,@attr)
                  ,(length *footnotes*)))))

(defmethod html(stream (tag (eql 'footnotes)) &optional attr content)
  (declare (ignore content))
  (when *footnotes*
    (dolist (item (getf attr :before '(((hr :align :left :width 0.3)))))
      (html stream item))
    (html stream
          `((ol :class :footnotes :id "footnotes")
            ,@(mapcar #'(lambda(footnote) `(li ,footnote))
                      (reverse *footnotes*))))
    (dolist (item (getf attr :after)) (html stream item))
    (setq *footnotes* nil)))