;;; -*- lisp -*-
;;;; $Id: latex.lisp,v 1.3 2007/07/16 07:29:48 willijar Exp willijar $
;;;; Copyright (C) 2003 Dr. John A.R. Williams, J.A.R.Williams@blueyonder.co.uk

;;;; This file is part of the Common Lisp Markup Package

;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.

;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.

;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
;;;; or retreive a copy from http://www.gnu.org/licenses/gpl.txt

;;;; Process functions for converting s-exp markup to latex

(in-package :markup)

(defvar *latex-class* "article")
(defvar *latex-options* '("10pt" "a4paper"))
(defvar *latex-packages* '("graphicx" "hyperref" "amsmath"))
(defvar *latex-prefix*
  '("\\setlength{\\topmargin}{1cm}"
    "\\setlength{\\oddsidemargin}{0cm}"
    "\\setlength{\\textwidth}{16cm}"
    "\\setlength{\\textheight}{22cm}"
    "\\newcommand{\\erfc}{\\rm{erfc}}"
    "\\newcommand{\\erf}{\\rm{erf}}"
    "\\newcommand{\\sinc}{\\rm{sinc}}"))

(defvar *escape-latex* t
  "If true latex special characters will be escaped")

(defvar *document-title* "")
(defvar *document-author* "")
(defvar *document-date* nil)
(defvar *no-columns* 1)

(defun latex-reserved-p (c) (member c '(#\$ #\& #\^ #\_ #\%)))
(defun latex-math-p(c) (member c '(#\> #\<)))
(defun latex-escape (stream s)
  (loop
     :for c :across s
     :when (latex-math-p c)
     :do (format stream "$~A$" c)
     :else
     :when (latex-reserved-p c) :do (write-char #\\ stream) :end
     :and :do (write-char c stream)
     :end))

(defun latex-escaped(s)
  (with-output-to-string(os)
    (latex-escape os
                  (if (stringp s) s (format nil "~A" s)))))


(defmacro with-latex-environment((stream name &optional options) &body body)
  (let ((gname (gensym)))
    `(let ((,gname ,name))
      (format ,stream "~%\\begin{~A}~@[~A~]~%" ,gname ,options)
      ,@body
      (format ,stream"~%\\end{~A}~%" ,gname))))

(defgeneric latex(stream tag &optional attr content)
  (:documentation "Convert markup to html"))

(defmethod latex((path pathname) markup  &optional attr content)
  (with-open-file(os path
                     :direction :output
                     :if-does-not-exist :create
                     :if-exists :supersede)
    (call-next-method os markup attr content)))

(defmethod latex(stream tag &optional attr content)
  (if (and (symbolp tag) (not (eql (symbol-package tag) (find-package :markup))))
      (latex stream (intern (symbol-name tag) (find-package :markup)) attr content)
      (progn
        (warn "Latex conversion for ~A not defined : content~%~A" tag content)
        (unless (empty-element-p tag)
          (dolist (item content) (latex stream item))))))

(defmethod latex(stream (tag number) &optional attr content)
  "Print numbers our"
  (declare (ignore attr content))
  (princ tag stream))

(defmethod latex(stream (markup list) &optional attr content)
  (declare (ignore content attr))
  (apply #'latex (cons stream (multiple-value-list (split-markup markup)))))

(defmethod latex((stream (eql nil)) markup &optional attr content)
  (with-output-to-string(s) (latex s markup attr content)))

(defmethod latex(stream (tag string) &optional attr content)
  (declare (ignore content attr))
  (if *escape-latex* (latex-escape stream tag)
      (write-sequence tag stream)))

(defmethod latex :around (stream (tag symbol) &optional attr content)
  (unless (getf attr :not-latex)
    (let ((*escape-latex* (getf attr :escape *escape-latex*)))
      (call-next-method stream tag attr content) )))

(defmethod latex(stream (tag (eql 'nil)) &optional attr content)
  (declare (ignore attr content))
  (write-char #\space stream))

(defmethod latex(stream (tag (eql 'html)) &optional attr content)
  (latex stream 'document attr content))

(defmethod latex(stream (tag (eql 'document)) &optional attr content)
  (let* ((*latex-options* (getf attr :options *latex-options*))
         (*latex-class* (getf attr :class *latex-class*))
         (*latex-packages* (getf attr :packages *latex-packages*))
         (*latex-prefix* (getf attr :prefix *latex-prefix*))
         (*no-columns*
          (if (member "twocolumn" *latex-options* :test #'string=) 2 1))
         (*section-level* (if (or (string= *latex-class* "book")
                                  (string= *latex-class* "report"))
                              -1 0)))
    (format stream "\\documentclass~:[~*~;[~:*~A~{,~A~}]~]{~A}~%~{\\usepackage{~A}~%~}~{~A~%~}"
            (first *latex-options*)
            (rest *latex-options*)
            *latex-class*
            *latex-packages*
            *latex-prefix*)
    (dolist (item content) (latex stream item))))

(defmethod latex(stream (tag (eql 'sup)) &optional attr content)
  (declare (ignore attr))
  (write-string "$^{" stream)
  (dolist (item content) (latex stream item))
  (write-string "}$" stream))

(defmethod latex(stream (tag (eql 'sub)) &optional attr content)
  (declare (ignore attr))
  (write-string "$_{" stream)
  (dolist (item content) (latex stream item))
  (write-string "}$" stream))

(defmethod latex(stream (tag (eql 'head)) &optional attr content)
  (declare (ignore attr))
  (dolist (item content) (latex stream item)))

(defmethod latex(stream (tag (eql 'title)) &optional attr content)
  (declare (ignore attr stream))
  (setq *document-title* (first content)))

(defmethod latex(stream (tag (eql 'author)) &optional attr content)
  (declare (ignore attr stream))
  (setq *document-author* (first content)))

(defmethod latex(stream (tag (eql 'date)) &optional attr content)
  (declare (ignore attr stream))
  (setq *document-date* (first content)))

(defmethod latex(stream (tag (eql 'toc)) &optional attr content)
  (declare (ignore attr content))
  (write-line "\\tableofcontents" stream))

(defmethod latex(stream (tag (eql 'abstract)) &optional attr content)
  (declare (ignore attr))
  (with-latex-environment(stream "abstract")
    (dolist (item content) (latex stream item))))

(defmethod latex(stream (tag (eql 'body)) &optional attr content)
  (declare (ignore attr))
  (with-latex-environment(stream "document")
    (format stream "\\title{~A}
\\author{~@[~A~]}
~@[\\date{~A}~]
\\maketitle"
            *document-title* *document-author* *document-date*)
    (dolist (item content) (latex stream item))))

(defmethod latex(stream (tag (eql 'p)) &optional attr content)
  (declare (ignore attr))
  (write-char #\newline stream)
  (dolist (item content) (latex stream item))
  (write-char #\newline stream))

(defmethod latex(stream (tag (eql 'a)) &optional attr content)
  (format stream "\\href{~A}{~A}"
          (getf attr :href)
          (latex nil (first content))))

(defvar *latex-sections*
  #("chapter" "section" "subsection" "subsubsection" "paragraph"
    "subparagraph")
  "Latex sections in depth order")

(defmethod latex(stream (tag (eql 'section)) &optional attrs content)
  (let ((*section-level* (or (getf attrs :level) (1+ *section-level*)))
        (title (getf attrs :title))
        (id (getf attrs :id)))
    (if  (< *section-level* (length *latex-sections*))
         (format stream "~%\\~A~:[~;*~]{~A}~%"
                 (elt *latex-sections* *section-level*)
                 (getf attrs :nonumber)
                 title)
         (format stream "~%\\em{~A}~%" title))
    (if id (format stream "\\label{~A}~%" id))
    (dolist (item content) (latex stream item))
    (write-char #\newline stream)))

(defun write-latex-header(stream attrs content)
  (format stream "~%\\~A{~A}~%"
          (elt *latex-sections* *section-level*)
          (first content))
  (let ((id (getf attrs :id)))
    (when id (format stream "\\label{~A}~%" id))))

(defmethod latex(stream (tag (eql 'h1)) &optional attrs content)
  (setq *section-level* 0)
  (write-latex-header stream attrs content))
(defmethod latex(stream (tag (eql 'h2)) &optional attrs content)
  (setq *section-level* 1)
  (write-latex-header stream attrs content))
(defmethod latex(stream (tag (eql 'h3)) &optional attrs content)
  (setq *section-level* 2)
  (write-latex-header stream attrs content))
(defmethod latex(stream (tag (eql 'h4)) &optional attrs content)
  (setq *section-level* 3)
  (write-latex-header stream attrs content))

(defmethod latex(stream (tag (eql 'div)) &optional attrs content)
  (declare (ignore attrs))
  (dolist (item content) (latex stream item)))

(defmethod latex(stream (tag (eql 'navbar)) &optional attrs content)
  (declare (ignore stream attrs content)))

(defmethod latex(stream (tag (eql 'math)) &optional attr content)
  (declare (ignore attr))
  (let ((math (string-trim '(#\Space #\Tab #\Newline) (first content))))
    (if (eq (elt math 0) #\$)
        (write-sequence (string-trim '(#\Space #\Tab #\Newline) math) stream)
        (format stream "$~A$" math))))

(defmethod latex(stream (tag (eql 'equation)) &optional attr content)
  (let((id (getf attr :id)))
    (let* ((math (string-trim '(#\Space #\Tab #\Newline) (first content))))
      (if (getf attr :raw)
          (write-sequence  math stream)
          (with-latex-environment
              (stream (format nil
                              "~:[~:[displaymath~;equation~]~;eqnarray~:[*~;~]~]"
                              (is-eqnarray-p math) id))
            (write-sequence  math stream)
            (if id (format stream "~%\\label{~A}" id)))))))

(defmethod latex(stream (tag (eql 'figure)) &optional attrs content)
  (let* ((max-width           ; max width specified for contained item
          (labels ((max-width(markup)
                     (multiple-value-bind(tag attr contents)
                         (split-markup markup)
                       (declare (ignore tag))
                       (if contents
                           (max (getf attr :width 0)
                                (reduce #'max (mapcar #'max-width contents)))
                           (getf attr :width 0)))))
            (max (getf attrs :width 0)
                 (reduce #'max (mapcar #'max-width content)))))
         (dbl  (> max-width (/ 1 *no-columns*)))
         (*no-columns* (if dbl 1 *no-columns*)))
    (with-latex-environment(stream (if dbl "figure*" "figure"))
      (format stream "\\center~%")
      (dolist (item content) (latex stream item))
      (let ((caption (getf attrs :caption))
            (id  (getf attrs :id)))
        (when caption (format stream "\\caption{~A}" caption))
        (when id (format stream "\\label{~A}~%" id))))))

(defun table-size(table-rows)
  "Returns no-of-columns,no-of-rows and widths given table content"
  (let ((no-of-rows 0)
        (no-of-columns 0)
        (widths (make-array 1 :initial-element nil :adjustable t)))
    (dolist (row table-rows)
      (multiple-value-bind (tag attr cols) (split-markup row)
        (declare (ignore attr))
        (when (eq 'tr tag)              ;  a valid row
          (incf no-of-rows)
          (let ((local-no-cols 0))
            (dolist (col cols)
              (multiple-value-bind (tag attr) (split-markup col)
                (when (or (eq tag 'td) (eq tag 'th)) ; a valid column
                  (let ((width (getf attr :width))
                        (colspan (getf attr :colspan 1)))
                    (incf local-no-cols colspan)
                    (when (> local-no-cols no-of-columns)
                      (setq no-of-columns local-no-cols)
                      (adjust-array widths no-of-columns :initial-element nil))
                    (when (and (= 1 colspan) width
                               (not (aref widths (- no-of-columns 1))))
                      (setf (aref widths (- no-of-columns 1)) width)))
                  )))
            ))))
    (values no-of-columns no-of-rows widths)))

(defun write-table-colspec(stream count align border widths)
  (when border (write-char #\| stream))
  (let ((align-char (cdr (assoc (string align)
                                '(("left" . #\l) ("right" . #\r)
                                  ("center" . #\c)) :test #'string-equal))))
    (dotimes(x count)
      (let ((width (aref widths x)))
        (if (numberp width)
            (format stream "p{~A\\textwidth}"
                    (if (<= width 1) width (/ width 100.0)))
            (write-char align-char stream)))
      (when border (write-char #\| stream)))))

(defmethod latex(stream (tag (eql 'table)) &optional attrs content)
  (let ((border (getf attrs :border)))
    (multiple-value-bind (no-cols no-rows widths) (table-size content)
      (declare (ignore no-rows))
      (terpri stream)
      (write-sequence "\\begin{tabular}{"  stream)
      (write-table-colspec stream no-cols :left border widths)
      (write-line "}" stream)
      (when border (write-line "\\hline\\hline" stream))
      (dolist (row content)
        (multiple-value-bind (tag attr cols) (split-markup row)
          (declare (ignore attr))
          (when (eq 'tr tag)
            (dolist (col cols)
              (unless (eq col (first cols)) (write-sequence " & " stream))
              (multiple-value-bind (tag attr content) (split-markup col)
                (when (or (eq tag 'td) (eq tag 'th))
                  (let* ((align (getf attr :align
                                      (if (eq tag 'td) :left :center)))
                         (cols (getf attr :colspan 1))
                         (width (make-array
                                 1
                                 :initial-element (list (getf attr :width)))))
                    (if (and (eq align :left) (eq cols 1))
                        (dolist (item content) (latex stream item))
                        (progn
                          (format stream "\\multicolumn{~D}{" cols)
                          (write-table-colspec stream 1 align border width)
                          (write-sequence "}{" stream)
                          (dolist (item content) (latex stream item))
                          (write-char #\} stream)))))))
            (write-sequence "\\\\" stream)
            (when border (write-sequence "\\hline" stream))
            (terpri stream))))
      (when border (write-sequence "\\hline" stream))
      (write-line "\\end{tabular}"  stream)
      (let ((caption (getf attrs :caption))
            (id  (getf attrs :id)))
        (when caption (format stream "\\caption{~A}~%" caption))
        (when id (format stream "\\label{~A}~%" id))))))

(defmethod latex(stream (tag (eql 'ref)) &optional attrs content)
  (let* ((id (first content))
         (ref (gethash id *references*)))
    (if (and ref (getf attrs :inplace) (fourth ref))
        (latex stream (fourth ref))
        (multiple-value-bind(tag) (split-markup (fourth ref))
          (if (or (eq tag 'equation) (eq tag 'figure))
              (format stream "~~\\ref{~A})" id)
              (format stream "~A (See~~\\ref{~A}) "
                      (or (second content) (first ref))
                      id))))))

(defmethod latex(stream (tag (eql 'b)) &optional attr content)
  (declare (ignore attr))
  (write-sequence "{\\bf " stream)
  (dolist (item content) (latex stream item))
  (write-char #\} stream))

(defmethod latex(stream (tag (eql 'strong)) &optional attr content)
  (latex stream 'b attr content))

(defmethod latex(stream (tag (eql 'code)) &optional attrs content)
	(declare (ignore attrs))
	(dolist (item content)
		(write-string "\\verb\"" stream)
		(write-string item stream)
		(write-string "\"" stream)))

(defmethod latex(stream (tag (eql 'pre)) &optional attrs content)
	(declare (ignore attrs))
  (with-latex-environment(stream "verbatim")
		(write-string (first content) stream)))

(defmethod latex(stream (tag (eql 'em)) &optional attr content)
  (declare (ignore attr))
  (write-sequence "{\\em " stream)
  (dolist (item content) (latex stream item))
  (write-char #\} stream))

(defmethod latex(stream (tag (eql 'br)) &optional attr content)
  (declare (ignore attr content))
  (write-line "\\\\" stream))

(defmethod latex(stream (tag (eql 'ul)) &optional attr content)
  (declare (ignore attr))
  (with-latex-environment(stream "itemize")
    (dolist (item content) (latex stream item))))

(defmethod latex(stream (tag (eql 'ol)) &optional attr content)
  (with-latex-environment(stream "enumerate")
    (when (getf attr :start)
      (format stream "\\setcounter{enumi}{~D}" (- (getf attr :start) 1)))
    (dolist (item content) (latex stream item))))

(defmethod latex(stream (tag (eql 'dl)) &optional attr content)
  (declare (ignore attr))
  (with-latex-environment(stream "description")
    (dolist (item content) (latex stream item))))

(defmethod latex(stream (tag (eql 'li)) &optional attr content)
  (write-sequence "\\item " stream)
  (dolist (item content) (latex stream item))
  (when-bind (id (getf attr :id)) (format stream "\\label{~A}" id))
  (write-char #\newline stream))

(defmethod latex(stream (tag (eql 'dt)) &optional attr content)
  (declare (ignore attr))
  (terpri stream)
  (write-sequence "\\item[" stream)
  (dolist (item content) (latex stream item))
  (write-sequence "] "stream)
  (terpri stream))

(defmethod latex(stream (tag (eql 'span)) &optional attr content)
  (declare (ignore attr))
  (dolist (item content) (latex stream item)) )

(defmethod latex(stream (tag (eql 'dd)) &optional attr content)
  (declare (ignore attr))
  (terpri stream)
  (dolist (item content) (latex stream item))
  (terpri stream))

(defmethod latex(stream (tag (eql 'quote)) &optional attr content)
  (declare (ignore attr))
  (with-latex-environment(stream "quote")
    (dolist (item content) (latex stream item))))

(defun latex-form-value(stream value &optional (min-size 5))
  (declare (ignore min-size))
;  (let ((v (with-output-to-string(s) (princ value s))))
;    (format stream "\\framebox[~Dem]{~A}" (max (length v) min-size) v)))
  (write-string (latex-escaped value) stream))

(defmethod latex(stream (tag (eql 'input)) &optional attr content)
  (declare (ignore content))
  (unless (eql (getf attr :type) :submit)
    (latex-form-value stream (getf attr :value) (getf attr :size 5))))

(defmethod latex(stream (tag (eql 'mcq)) &optional attr content)
  (let ((value (getf attr :value)))
    (format stream "~%~%\\begin{minipage}{\\linewidth}{~%\\begin{tabular}{rp{0.85\\linewidth}}
~{$~:[\\circ~;\\bullet~]$ & ~A\\\\~%~}\\end{tabular}}\\end{minipage}"
            (mapcan
             #'(lambda(item)
                 (if (consp item)
                     (list (equal (car item) value) (latex-escaped (cdr item)))
                     (list (equal item value) (latex-escaped item))))
             content))))

(defmethod latex(stream (tag (eql 'boolean))  &optional attr content)
  (declare (ignore attr content))
  (latex stream 'mcq nil '("Yes" "No")))

(defmethod latex(stream (tag (eql 'maq)) &optional attr content)
  (declare (ignore content))
  (latex-form-value stream (getf attr :value) (getf attr :size 5)))

(defmethod latex(stream (tag (eql 'textarea)) &optional attrs content)
  (let ((value (getf attrs :value content)))
    (format stream "~%~%\\framebox{\\begin{minipage}[t]{\\textwidth}~%")
    (if value
        (progn
          (terpri stream)
          (latex-escape stream value)
          (terpri stream))
;        (if (getf attrs :structured-text)
;            (format stream "\par~%
;            (latex stream (list 'div (latex-value))
;            (latex stream value))
        (format stream "\\vspace{~Dex}" (getf attrs :cols 10)))
    (format stream "~%\\end{minipage}}")))

(defmethod latex(stream (tag (eql 'form)) &optional attr content)
  (declare (ignore attr))
  (terpri stream)
  (dolist (item content) (latex stream item))
  (terpri stream))

(defun latex-media-path(attrs source)
  (let* ((media-key
          (or (getf attrs :media)
              (getf (preprocess-media attrs (pathname source)
                                      :output "image/x-eps")
                    :media))))
    (media-variant-path
     (media-variant media-key *media-server* "image/x-eps")
     *media-server*)))

(defmethod latex(stream (tag (eql 'graphic)) &optional attr content)
  (format stream "\\includegraphics[width=~,2f\\textwidth,clip=true]{~A}~%"
          (let ((width (getf attr :width 1.0)))
            (if (< width (/ 1 *no-columns*))
                width
                (/ 1 *no-columns*)))
          (latex-media-path attr (getf attr :src (first content)))))

(defmethod latex(stream (tag (eql 'footnote)) &optional attr content)
  (declare (ignore attr))
  (write-sequence "\\footnote{" stream)
  (latex-escape stream (first content))
  (write-char #\} stream))

(defmethod latex(stream (tag (eql'tableofcontents))
                 &optional attr content)
  (declare (ignore attr content))
  (write-line "\\tableofcontents" stream))

(defmethod latex(stream (tag (eql 'structured-text))
                 &optional attrs content)
  "Context is restored as structured text elements won't have
attributes from preprocessing"
  (let ((*figure-counter*
         (getf attrs :figure-counter *figure-counter*))
        (*equation-counter*
         (getf attrs :equation-counter *equation-counter*))
        (*document-name* (getf attrs :document-name *document-name*))
        (*file-counter* (getf attrs :file-counter *file-counter*)))
    (dolist (item content)
      (latex stream (parse-structured-text item)))))

(defgeneric pdf(os element)
  (:documentation "Write element as a pdf file to stream os"))

(defmethod pdf(out-stream (markup list))
  "Write give markup to stream - uses latex processing"
  (pdf out-stream (with-output-to-string(os) (latex os markup))))

(defmethod pdf(out (tex string))
    "Create a pdf from a tex string"
    (jarw.io:with-temp-pathname(texfile :type "tex")
      (let ((pdffile (merge-pathnames (make-pathname :type "pdf")
                                      texfile)))
        (with-open-file(os texfile
                           :direction :output
                           :if-exists :overwrite
                           :if-does-not-exist :create)
          (write-string tex os))
        (jarw.media::convert-media "application/x-latex" "application/pdf"
                                   texfile pdffile)
        (with-open-file(is pdffile
                           :direction :input
                           :element-type 'flexi-streams:octet)
          (let ((count (file-length is)))
            (cond
              ((streamp out) (jarw.io::copy-stream is out) out)
              ((and (symbolp out) (subtypep out 'flexi-streams:octet))
               (let ((sequence (make-array count :element-type out)))
                 (read-sequence sequence is)
                 sequence))
              ((typep out 'vector)
               (adjust-array out count :element-type 'flexi-streams:octet)
               (read-sequence out is :start 0 :end count)
               out)
              (t (error "Unknown type for PDF"))))))))


