;;;; $Id: markup.lisp,v 1.3 2007/01/21 08:31:06 willijar Exp $
;;;; Copyright (C) 2002 Dr. John A.R. Williams, J.A.R.Williams@blueyonder.co.uk
;;;; This file is part of the Common Lisp Markup Package

;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.

;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.

;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
;;;; or retreive a copy from http://www.gnu.org/licenses/gpl.txt

(in-package :markup)

(declaim (inline split-markup join-markup process-markup))
(defun split-markup(things)
  "Returns the tag, attributes and contents of an s-exp"
  (declare (optimize (speed 3)))
  (cond ((atom things) (values things))
        ((consp (car things))
         (values (caar things) (cdar things) (cdr things)))
        (t (values (car things) () (cdr things)))))

(defun join-markup(tag attrs content)
  "Given tag, attributes and content constructs an s-exp"
  #+cmu(declare (optimize (extensions:inhibit-warnings 3) (speed 3)))
  (if attrs
      (cons (cons tag attrs) (if (listp content) content (list content)))
      (cons tag (if (listp content) content (list content)))))

(defun process-markup (things func)
  (declare (optimize speed) (list things) (function func))
  (apply func (multiple-value-list (split-markup things))))

(declaim (inline markup-tag markup-attr markup-content))
(defun markup-tag(things)
  (declare (optimize (speed 3)))
  (cond ((atom things) things)
        ((consp (car things)) (caar things))
        (t (car things))))

(defun markup-attr(things field)
  (declare (optimize (speed 3)))
  (when (and (consp things) (consp (car things)))
    (getf (cdar things) field)))

(defun markup-content(things)
  (when (listp things) (cdr things)))

(defun substitute-markup-when (condition things function &key as-list)
  "Recurses through things, passing them to condition.
    If condition returns true function is called"
  (declare (function condition) (list things) (function function))
  (if (funcall condition things)
      (funcall function things)
      (if as-list
          (list
           (cons (first things)
                 (mapcan
                  #'(lambda(e)
                      (if (listp e)
                          (substitute-markup-when condition e function
                                                  :as-list as-list)
                          (list e)))
                  (rest things))))
          (cons (first things)
                (mapcar
                 #'(lambda(e)
                     (if (listp e)
                         (substitute-markup-when condition e function
                                                 :as-list as-list)
                         e))
                 (rest things))))))

(defun process-markup-when (condition things function)
  "Recurses through things, passing them to condition.
   If condition returns true function is called with things"
  (when (and function (or (eq 't condition) (funcall condition things)))
    (funcall function things))
  (when (and (listp things) (listp (rest things)))
    (dolist (more-things (rest things))
      (process-markup-when condition more-things function))))

(defgeneric title(item)
  (:documentation "Return the title of an item"))

(defmethod title((markup list))
  "Returns the topmost title of this piece of markup"
  (process-markup-when
   t markup
   #'(lambda(m)
       (multiple-value-bind(tag attrs content)(split-markup m)
         (when-bind(title
                    (cond
                      ((eq tag 'title) (first content))
                      ((getf attrs :title))))
           (return-from title title))))))

(defun remove-attribute(attribute list &key (test #'eq))
  "Returns a NEW attribute list with the specified attribute removed"
  (loop for a on list by #'cddr
        unless (funcall test attribute (car a))
        append (list (car a) (cadr a))))

(defun remove-attribute-if(test list)
  "Returns a NEW attribute list with the specified attribute removed"
  (loop for a on list by #'cddr
        unless (funcall test (car a))
        append (list (car a) (cadr a))))

(defmacro with-markup-environment ((&rest rest &key &allow-other-keys)
                                   &body body)
  "Establish an environment for preprocessing or rendering markup. Binds
   special variables according to specified keys. The meaning of keys is
   processing specific."
  (let ((vars '())
        (defs '()))
    (while rest
      (let ((var (intern (concatenate 'string
                                      "*" (string (first rest)) "*")
                         :markup ))
            (value (second rest)))
        (push var vars)
        (push (list var value) defs)
        (setq rest (rest (rest rest)))))
    `(let (,@defs)
      (declare (special ,@vars))
      ,@body)))


(defvar *textual-attributes* '(:title :caption)
  "A list of attributes which contain text - used when extracting keywords")

(defun markup-keywords(markup &key (ignore-word-p #'ignore-word-p))
  "Returns a list of keywords extracted from given markup"
  (let ((markup-strings '()))
    (process-markup-when
     't markup
     #'(lambda(things)
         (multiple-value-bind (tag attr) (split-markup things)
           (when (stringp tag) (push tag markup-strings))
           (when attr
             (dolist (a *textual-attributes*)
               (let ((s (getf attr a)))
                 (when s (push (if (listp s)
                                   (markup-keywords
                                    (cons 'span s)
                                    :ignore-word-p ignore-word-p)
                                   s)
                               markup-strings))))))))
    (mapcan
     #'(lambda(text) (keywords text :ignore-word-p ignore-word-p))
     markup-strings)))

(defmacro markup-transformation((tagvar attrvar contentvar &optional markupvar)
                                &body body)
  "Macro returns a function to transform a markup tree.  During
execution of the body the tagvar, attrvar and contentvar are bound to
the tag, attributes and content list of the current markup element at
the start.If markupvar is provided it will be bound to the original
markup cons.

The body is a list of clauses consisting of a test form and form*
return the new markup and (optionally) a value of t if the
transformation is to be re-applied to the markup content.

The test form may either be a symbol which is tested against tagvar or
a general test form."
  (let ((case-p (every #'symbolp (mapcar #'car body)))
        (func (gensym))
        (markup (or markupvar (gensym)))
        (descent (gensym)))
    `(lambda(,markup)
      (labels((,func(,markup)
                (multiple-value-bind(,tagvar ,attrvar ,contentvar)
                    (split-markup ,markup)
                  (multiple-value-bind(,markup ,descent)
                      (,@(if case-p `(case ,tagvar) '(cond))
                         ,@(mapcar
                            #'(lambda(clause)
                                (if (or case-p (not (symbolp (car clause))))
                                    clause
                                    (cons `(eql ,tagvar ',(car clause))
                                          (cdr clause))))
                            body))
                    (if ,markup
                        (if ,descent
                            (multiple-value-bind(,tagvar ,attrvar ,contentvar)
                                (split-markup ,markup)
                              (join-markup ,tagvar ,attrvar
                                           (mapcar #',func ,contentvar)))
                            ,markup)
                        (join-markup ,tagvar ,attrvar
                                     (mapcar #',func ,contentvar)))))))
        (,func ,markup)))))

(defmacro markup-scanner((tagvar attrvar contentvar &optional markupvar)
                         &body body)
  "Macro returns a function to scan a markup tree.  During execution
of the body the tagvar, attrvar and contentvar are bound to the tag,
attributes and content list of the current markup element at the
start.

The body is a list of clauses consisting of a test form and form*
which are executed for side effect and return a value of t if the
scan is to be reapplied to the markup content.

The test form may either be a symbol which is tested against tagvar or
a general test form."
  (let ((case-p (every #'symbolp (mapcar #'car body)))
        (func (gensym))
        (markup (or markupvar (gensym))))
    `(lambda(,markup)
      (labels((,func(,markup)
                (multiple-value-bind(,tagvar ,attrvar ,contentvar)
                    (split-markup ,markup)
                  (when
                      (,@(if case-p `(case ,tagvar) '(cond))
                         ,@(mapcar
                            #'(lambda(clause)
                                (if (or case-p (not (symbolp (car clause))))
                                    clause
                                    (cons `(eql ,tagvar ',(car clause))
                                          (cdr clause))))
                            body)
                         (t))
                    (map 'nil #',func ,contentvar)))))
        (,func ,markup)))))