;;;; $Id: regex.lisp,v 1.1 2006/07/18 18:52:36 willijar Exp $
;;;; Extensions to regex library
;;;; Copyright (C) 2004-2006 Dr. John A.R. Williams
;;;;
;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or (at
;;;; your option) any later version.
;;;;
;;;; This program is distributed in the hope that it will be useful, but
;;;; WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;;; General Public License for more details.
;;;;
;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
;;;; 02110-1301, USA.;;;; REGEX stuff

(in-package :regex)

(eval-when(:compile-toplevel :load-toplevel)
(export '(scan-str-multiple
          regex-split-string
          regex-process-string
          compile-regex-substitution
          regex-substitute-string
          regex-quote-string)))

(defun scan-str-multiple (matcher candstr
                          &key (start 0)
                          (end (length candstr))
                          (max-matches nil)
                          (match-register 0))
  "Run a matcher against a candidate string, scanning forward if necessary.
   Multiple matches are detected, and the registers of each match are returned
   (in reverse order) as a list. returns nil on failure."
  (declare (type string candstr)
           (type fixnum start end  match-register))
  (declare (optimize speed))
  (let ((result '())
        (no-matches 0))
    (declare (type fixnum no-matches) (type list result))
    (loop
     (multiple-value-bind (matched left matchlen regs)
         (scan-str  matcher candstr :start start :length (- end start))
       (declare (ignore left matchlen) (type boolean matched)
                (type (simple-array list *) regs))
       (unless matched (return result))
       (push regs result)
       (setq start (cdr (elt regs match-register)))
       (when (and max-matches (>= (incf no-matches) no-matches))
         (return result) )))))


(let ((reg-matcher (compile-str "[^\\\\]\\\\([:digit:]+)")))
  (defun compile-regex-substitution(str)
    "Compile a substitution string into a substitution list for scan-str-substitute"
    (let ((substr '())
          (end (length str)))
      (dolist (regs (scan-str-multiple reg-matcher str :end end))
        (push (subseq str (cdr (elt regs 0)) end) substr)
        (setq end (1+ (car (elt regs 0))))
        (push (parse-integer
               (subseq str (car (elt regs 1)) (cdr (elt regs 0))))
              substr))
      (push (subseq str 0 end) substr)
      substr )))

(defun regex-process-string (matcher candstr func
                             &key (start 0)
                             (end (length candstr))
                             (max-matches nil)
                             (match-register 0))
  "Run a matcher against a candidate string, scanning forward if
necessary.  A list is returned consisting of the unmatched strings and
the results of applying the function func to the matched
components. func is passed an array of registers corresponding to the
stored substring pattersn from the match. match-register specifies
which register is to be used to identify the end of the match.,
Default is 0, the whole pattern however it may be set to identify a
subexpression in context e.g. surrounded by spaces"
  (let ((substr '()))               ; build up list of substrings here
    (dolist (regs (scan-str-multiple matcher candstr
                                     :start start :end end
                                     :max-matches max-matches
                                     :match-register match-register))
      (unless (= (cdr (elt regs 0)) end)
        (push (subseq candstr (cdr (elt regs match-register)) end) substr)) ; unmatched tail - match end
      (setq end (car (elt regs match-register))) ; match start
      (push (funcall func (map 'vector
                               #'(lambda(c) (subseq candstr (car c) (cdr c)))
                               regs)) substr))
    (unless (= 0 end) (push (subseq candstr 0 end) substr))
    substr))

(defun regex-split-string (matcher string &key
                           (remove-empty-substr nil)
                           (start 0)
                           (end nil)
                           (max-split nil))
  "Return a list of substrings in string delimited by regex matcher
. If :remove-empty-subseqs is NIL, empty subsequences will be included
in the result; otherwise they will be discarded (and are not counted
in the number of splits). :start and :end define the range indices of
the string to be parsed. if defined :max-substr If it is nonzero, at
most maxsplit number of splits occur, and the remainder of the string
is returned as the final element of the list (thus, the list will have
at most maxsplit+1 elements"
  (let ((end (if end (min end (length string)) (length string)))
        (left start)
        (no-split 0)
        (list '()))
    (nreverse
     (loop
      (when (>= left end) (return list))
      (when (and max-split (>= no-split max-split))
        (return (push (subseq string left end) list)))
      (multiple-value-bind (matched right matchlen)
          (scan-str  matcher string :start left :length (- end left))
        (let ((right (if matched right end)))
          (unless (and remove-empty-substr (= left right))
            (push (subseq string left right) list)
            (incf no-split)))
        (unless matched (return list))
        (setf left (+ right matchlen)) )) )))


(defun regex-substitute-string(matcher candstr substitution
                               &key (start 0)
                               (end (length candstr))
                               (max-matches nil)
                               (match-register 0))
  "Run a matcher against a candidate string, scanning forward if necessary.
   Replace matches with substitution sub which should be a list of
   strings, numbers or functions. Strings are substituted literally,numbers
   refer to the corresponding match register. Functions are passed the vector
   of pattern match registers"
  (apply
   #'concatenate
   (cons 'string
         (regex-process-string
          matcher candstr
          #'(lambda(regs)
              (apply #'concatenate
                     (cons 'string
                           (mapcar #'(lambda(s)
                                       (typecase s
                                         (number (elt regs s))
                                         (function (funcall s regs))
                                         (t s)))
                                   substitution))))
          :start start
          :end end
          :max-matches max-matches
          :match-register match-register))))

(defun regex-quote-string(string)
  "Quotes a string so it can be substituted as a literal
   into a regex expression"
  (regex-substitute-string
   (load-time-value
    (regex:compile-str "[\\\\{\\}\\[\\]\\(\\)\\|\\.\\*\\$\\?\\^]"))
   string
   '("\\" 0)))

