;;;; -*- lisp -*-
;;;; $Id: rst.lisp,v 1.3 2007/07/13 10:01:05 willijar Exp $
;;;; Copyright (C) 2002 Dr. John A.R. Williams, J.A.R.Williams@blueyonder.co.uk

;;;; This file is part of the Common Lisp Markup Package

;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.

;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.

;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
;;;; or retreive a copy from http://www.gnu.org/licenses/gpl.txt

;;;; Process functions for converting s-exp markup to html

(in-package :markup)

(defclass rst-writer()
  ((stream :reader os :initarg :stream :initform *standard-output*)
   (current-column :initform 0 :reader current-column)
   (current-indent :initform 0 :reader current-indent)
   (fill-column :initarg :fill-column :initform 72 :reader fill-column))
  (:documentation "Class for managing output of restructured text"))

(defun write-newline(writer)
  (terpri (os writer))
  (setf (slot-value writer 'current-column) 0))

(defun repeat-write(stream char repeat)
  (dotimes(x repeat) (write-char char stream)))

(defun write-indent(writer)
  (assert (= 0 (current-column writer)))
  (repeat-write (os writer) #\space (current-indent writer))
  (setf (slot-value writer 'current-column) (current-indent writer)))

(defmacro with-rst-block((writer &key (indent 0) fill)
                         &body body)
  (let ((gwriter (gensym))
        (gindent (gensym))
        (gfill (gensym)))
    `(let* ((,gwriter ,writer)
            (,gindent (current-indent ,gwriter))
            (,gfill (fill-column ,gwriter)))
      (unwind-protect
           (progn
             (setf (slot-value ,gwriter 'current-indent) (+ ,gindent ,indent))
             ,@(when fill `((setf (slot-value ,gwriter 'fill-column) ,fill)))
             ,@body
             (write-newline ,gwriter))
        (setf (slot-value ,gwriter 'current-indent) ,gindent)
        (setf (slot-value ,gwriter 'fill-column) ,gfill)))))

(defparameter +rst-token-delimiters+ '(#\space #\tab #\return #\linefeed #\~))
(defparameter +rst-title-styles+ #("==" "--" "=" "-" "+" "."))

(defun rst-tokens(string)
  (let ((tokens nil)
        (start 0))
    (loop
     (let ((p (position-if
               #'(lambda(c) (member c +rst-token-delimiters+))
               string
               :start start)))
       (cond
         (p
          (when (> p start) (push (subseq string start p) tokens))
          (setf start (1+ p)))
         (t
          (when (< start (length string))
            (push (subseq string start) tokens))
          (return (nreverse tokens))))))))

(defgeneric rst-out(object writer)
  (:documentation "Writer an object to an rst-writer"))

(defmethod rst-out(object (writer rst-writer))
  (rst-out (write-to-string  object :readably nil :escape nil)
           writer))

(defmethod rst-out((string string) (writer rst-writer))
  (let ((tokens (rst-tokens string)))
    (when tokens
      (dolist(token tokens)
        (when (and (> (current-column writer) (current-indent writer))
                   (> (+ 1 (current-column writer) (length token))
                      (fill-column writer)))
          (write-newline writer))
        (when (= 0 (current-column writer))
          (write-indent writer))
        (when (> (current-column writer) (current-indent writer))
          (write-char #\space (os writer))
          (incf (slot-value writer 'current-column)))
        (write-string token (os writer))
        (incf (slot-value writer 'current-column) (length token)))
      (when (member (char string (1- (length string)))
                    +rst-token-delimiters+)
        (write-char #\space (os writer))
        (incf (slot-value writer 'current-column))))))

(defgeneric rst(stream tag &optional attr content)
  (:documentation "Convert markup to rst - returns stream"))

(defmethod rst((stream stream) tag &optional attr content)
  (rst (make-instance 'rst-writer :stream stream) tag attr content))

(defmethod rst((writer (eql nil)) markup &optional attr content)
  (with-output-to-string(os)
    (let ((writer (make-instance 'rst-writer :stream os)))
      (rst writer markup attr content))))

(defmethod rst(writer (markup list)  &optional attr content)
  (declare (ignore attr content))
  (apply #'rst (cons writer (multiple-value-list (split-markup markup)))))

(defmethod rst(writer (tag string) &optional attr content)
  (declare (ignore content attr))
  (rst-out tag writer))

(defmethod rst(writer (tag number) &optional attr content)
  (declare (ignore content attr))
  (rst-out (write-to-string tag :escape nil :readably nil) writer))

(defmethod rst :around (writer (tag symbol) &optional attr content)
  (declare (ignore content))
  (when (block-element-p tag)
    (write-newline writer)
    (let ((id (getf attr :id)))
      (when id
        (rst-out (format nil ".. _~A:" (string-downcase id))  writer )
        (write-newline writer)
        (write-newline writer))))
  (call-next-method))

(defun rstc(writer content)
  (if writer
      (dolist(item content) (rst writer item))
      (with-output-to-string(os)
        (let ((writer (make-instance 'rst-writer :stream os)))
          (dolist(item content) writer (rst writer item))))))

(defmethod rst(writer (tag symbol) &optional attr content)
  (declare (ignore attr content))
  (error "Tag ~S not supported for rst output." tag))

(defmethod rst(writer (tag (eql 'html)) &optional attr content)
  (declare (ignore attr))
  (rstc writer content))

(defmethod rst(writer (tag (eql 'document)) &optional attr content)
  (rst writer 'html attr content))

(defmethod rst(writer (tag (eql 'body)) &optional attrs content)
	(declare (ignore attrs))
	(let ((*section-level* 0))
    (setf (slot-value writer 'current-indent) 0)
	  (rstc writer content)))

(defun rst-title(writer content)
  (let ((title (if (listp content) (rstc nil content) content))
        (style (elt +rst-title-styles+ *section-level*)))
    (when (> (length style) 1)
      (repeat-write (os writer) (elt style 0) (length title)))
    (write-newline writer)
    (rst-out title writer)
    (write-newline writer)
    (repeat-write (os writer)
                  (elt style (if  (> (length style) 1) 1 0))
                  (length title))
    (write-newline writer)))

(defmethod rst(writer (tag (eql 'title)) &optional attr content)
  (declare (ignore attr))
  (rst-title writer content))

(defun rst-field-item(writer item content)
  (rst-out (strcat ":" item ": ") writer)
  (with-rst-block(writer :indent (+ 3 (length item)))
    (rstc writer content)
    (write-newline writer)))

(defmethod rst(writer (tag (eql 'author)) &optional attr content)
  (declare (ignore attr))
  (rst-field-item writer "author" content))

(defmethod rst(writer (tag (eql 'date)) &optional attr content)
  (declare (ignore attr))
  (rst-field-item writer "date" content))

(defmethod rst(writer (tag (eql 'abstract)) &optional attr content)
  (declare (ignore attr))
  (rst-field-item writer "abstract" content))

(defmethod rst(writer (tag (eql 'section)) &optional attrs content)
  (let ((*section-level* (or (getf attrs :level) (1+ *section-level*)))
        (title (getf attrs :title)))
    (with-rst-block(writer)
      (rst-title writer title)
      (rstc writer content))))

(defmethod rst(writer (tag (eql 'p)) &optional attr content)
  (let ((class (getf attr :class)))
    (cond
      ((eql class :error)
       (rst-out ".. warning:: " writer)
       (write-newline writer)
       (with-rst-block(writer :indent 3) (rstc writer content)))
      (t
       (when class
         (rst-out (strcat ".. class:: " (string-downcase class)) writer)
         (write-newline writer))
       (with-rst-block(writer) (rstc writer content))))))

(defmethod rst (writer (tag (eql 'math)) &optional attr content)
  (declare (ignore attr))
  (rst-out  (format nil "`~A`:math:" (first content)) writer))

(defmethod rst (writer (tag (eql 'equation)) &optional attr content)
  (declare (ignore attr))
  (let ((math (first content)))
    (let ((mathtype (if (is-eqnarray-p math) "eqnarray*" "displaymath")))
      (rst-out ".. equation:: " writer)
      (with-rst-block(writer :indent 3)
        (write-newline writer)
        (rst-out (format nil "\\begin{~A}" mathtype) writer)
        (write-newline writer)
        (with-rst-block(writer :indent 2) (rst-out math writer))
        (rst-out (format nil "\\end{~A}" mathtype) writer)
        (write-newline writer)))))

(defun rst-fields-out(writer alist)
  (dolist(item alist)
    (let ((name (format nil "~A " (string (car item)))))
      (rst-out name writer)
      (when (cdr item)
        (with-rst-block(writer :indent (length name))
          (rst-out (cdr item) writer)))))
  (write-newline writer))

(defun parse-width(width)
  (unless (numberp width)
    (setf width (string width))
    (setf width (data-format-validation::parse-number
                 width
                 :end (position-if-not #'digit-char-p width
                                       :from-end t)))
    (if (< width 1) (* 100 width) width)))

(defmethod rst (writer (tag (eql 'figure)) &optional attrs content)
  (let ((caption (getf attrs :caption)))
    (if (eql (markup-tag (first content)) 'graphic)
        (let* ((graphic (first content))
               (align (getf attrs :align))
               (url (first (markup-content graphic)))
               (alt (markup-attr graphic :alt))
               (width (parse-width (markup-attr graphic :width))))
          (rst-out ".. figure::" writer)
          (with-rst-block(writer :indent 3)
            (rst-out url writer)
            (write-newline writer)
            (rst-fields-out
             writer
             `((":alt:" . ,(or alt caption url))
               ,@(when align `((":align:" . ,(string-downcase align))))
               ,@(when width `((":width:" . ,width)))))
            (when caption
              (rst-out caption writer)
              (write-newline writer))))
        (progn
          (rst-out (strcat ".. topic:: " caption) writer)
          (write-newline writer)
          (with-rst-block(writer :indent 3)
            (rstc writer content))))))

(defmethod rst (writer (tag (eql 'graphic)) &optional attrs content)
  (let ((align (getf attrs :align))
        (url (first content))
        (width (getf attrs :width))
        (alt (getf attrs :alt)))
    (rst-out ".. image:" writer)
    (with-rst-block(writer :indent 3)
      (rst-out url writer)
      (rst-fields-out
       writer
       `(,@(when align `(("align" . ,(string-downcase align))))
         ,@(when width `((width . ,width)))
         ("alt" . ,(or alt url)))))))

(defmethod rst (writer (tag (eql 'file)) &optional attrs content)
  (rst-out (format nil " `~A <~A>`_ " (first content)
                   (html-media-url attrs (pathname (first content))
                                   :output nil))
           writer))

(defmethod rst(writer (tag (eql 'ref)) &optional attrs content)
  (declare (ignore attrs))
  (let* ((id (first content))
         (ref (gethash id *references*))
         (title (first ref)))
    (rst-out (format nil "`~A`_" (or title (string-downcase id))) writer)))

(defmethod rst(writer (tag (eql 'cite)) &optional attr content)
  (declare (ignore attr))
  (rst-out (format nil "[~A]_" (first content)) writer))

(defmethod rst(writer (tag (eql 'footnote)) &optional attr content)
  (declare (ignore attr))
  (push (first content) *footnotes*)
  (rst-out "[~D]_" (length *footnotes*)))

(defmethod rst(writer (tag (eql 'footnotes)) &optional attr content)
  (declare (ignore content attr))
  (let ((c 0))
    (dolist(footnote *footnotes*)
      (rst-out (format nil ".. [~D] " (incf c)) writer)
      (with-rst-block(writer :indent 3)
        (rst-out footnote writer)))))

(defmethod rst(writer (tag (eql 'toc)) &optional attr content)
  (declare (ignore attr content))
  (write-newline writer)
  (rst-out ".. contents::" writer)
  (write-newline writer))

(defmethod rst(writer (tag (eql 'a)) &optional attr content)
  (rst-out (format nil "`~A <~A>`_" (first content) (getf attr :href))
           writer))

(defmethod rst(writer (tag (eql 'em)) &optional attr content)
	(declare (ignore attr))
  (rst-out (format nil "*~A*" (rstc nil content)) writer))

(defmethod rst(writer (tag (eql 'b)) &optional attr content)
	(declare (ignore attr))
  (rst-out (format nil "**~A**" (rstc nil content)) writer))

(defmethod rst(writer (tag (eql 'strong)) &optional attr content)
  (rst writer 'b attr content))

(defmethod rst(writer (tag (eql 'code)) &optional attrs content)
	(declare (ignore attrs))
  (rst-out (format nil "``~A``" (rstc nil content)) writer))

(defmethod rst(writer (tag (eql 'pre)) &optional attrs content)
	(declare (ignore attrs))
  (with-input-from-string(is (first content))
    (do((line (read-line is nil) (read-line nil)))
       ((not line))
      (write-indent writer)
      (write-string "| " (os writer))
      (write-string line (os writer))
      (write-newline writer))))

(defmethod rst(writer (tag (eql 'structured-text))  &optional attrs content)
  (rst writer 'pre attrs content))

(defmethod rst(writer (tag (eql 'h1)) &optional attrs content)
  (declare (ignore attrs))
  (setq *section-level* 0)
  (rst-title writer content))
(defmethod rst(writer (tag (eql 'h2)) &optional attrs content)
  (declare (ignore attrs))
  (setq *section-level* 1)
  (rst-title writer content))
(defmethod rst(writer (tag (eql 'h3)) &optional attrs content)
  (declare (ignore attrs))
  (setq *section-level* 2)
  (rst-title writer content))
(defmethod rst(writer (tag (eql 'h4)) &optional attrs content)
  (declare (ignore attrs))
  (setq *section-level* 3)
  (rst-title writer content))

(defmethod rst(writer (tag (eql 'div)) &optional attrs content)
  (declare (ignore attrs))
  (write-newline writer)
  (rstc writer content))

(defmethod rst(writer (tag (eql 'br)) &optional attr content)
  (declare (ignore attr content))
  (write-newline writer))

(defmethod rst(writer (tag (eql 'ul)) &optional attr content)
  (declare (ignore attr))
  (dolist(item content)
    (rst-out "* " writer)
    (with-rst-block(writer :indent 2)
      (rstc writer (cdr item)))))

(defmethod rst(writer (tag (eql 'ol)) &optional attr content)
  (declare (ignore attr))
  (let ((c 0))
    (dolist(item content)
      (rst-out (format nil "~D. " (incf c)) writer)
      (with-rst-block(writer :indent 3)
        (rstc writer (cdr item))))))

(defmethod rst(writer (tag (eql 'dl)) &optional attr content)
  (declare (ignore attr))
  (let ((indent 0))
    (dolist(item content)
      (multiple-value-bind(tag attr content) (split-markup item)
        (declare (ignore attr))
        (ecase tag
          (dt (let ((dt (strcat ":" (rstc nil content) ": ")))
                (setf indent (min (length dt) 4))
                (rst-out dt writer)))
          (dd (with-rst-block(writer :indent indent)
                (rstc writer content))))))))

(defmethod rst(writer (tag (eql 'span)) &optional attr content)
  (declare (ignore attr))
  (rstc writer content))

(defmethod rst(writer (tag (eql 'quote)) &optional attr content)
  (declare (ignore attr))
  (write-newline writer)
  (with-rst-block(writer :indent 4)
    (rstc writer content)))

(defmethod rst(writer (tag (eql 'table)) &optional attrs content)
  (declare (ignore attrs))
  (multiple-value-bind (no-cols no-rows widths) (table-size content)
    (dotimes(x (length widths))
      (let ((a (aref widths x)))
        (when a (setf (aref widths x) (parse-width a)))))
    ;; Set up widths as character widths
    (let* ((total-width (loop :for a :across widths :when a :summing a))
           (no-blank (loop :for a :across widths :count (not a)))
           (fixed-width (when (/= 0 no-blank)
                          (/ (- 100 total-width) no-blank))))
      (dotimes(x (length widths))
        (let ((width (aref widths x)))
          (setf (aref widths x) (* (/ (- 72 (* 2 no-cols)) 100) (or width fixed-width))))))
    ;; set up array of cell contents, colspans and whether is header row
    (let((contents (make-array (list no-rows no-cols) :initial-element nil))
         (colspans (make-array (list no-rows no-cols) :initial-element nil))
         (header? (make-array no-rows :element-type 'boolean
                              :initial-element nil))
         (heights (make-array no-rows :element-type 'fixnum
                              :initial-element 1)))
      (do*((row 0 (1+ row))
           (rest content (cdr rest))
           (tr (car rest) (car rest)))
          ((not tr))
        (setf (aref header? row) t)
        (do*((col 0 (1+ col))
             (rest (rest tr) (rest rest))
             (cell (car rest) (car rest)))
            ((not cell))
          (multiple-value-bind(tag attr content) (split-markup cell)
            (unless (eql tag 'th) (setf (aref header? row) nil))
            (setf (aref contents row col)
                  (split-sequence:split-sequence
                   #\newline
                   (with-output-to-string(os)
                     (rstc (make-instance 'rst-writer
                                          :stream os
                                          :fill-column (- (aref widths col) 2))
                           content))))
            (setf (aref heights row)
                  (max (aref heights row)
                       (length (aref contents row col))))
            (let ((colspan (getf attr :colspan)))
              (setf (aref colspans row col) (or colspan 1))
              (when colspan (incf col (1- colspan)))))))
      ;; Now actually write out table
      (labels((out(s) ;; write out a string with no wrapping
                (write-string s (os writer))
                (incf (slot-value writer 'current-column) (length s)))
              (tab(n &optional (char #\space)) ;; tabulate to start nth column
                (let ((col (+ (current-indent writer) (* n 2))))
                  (dotimes(x n) (incf col (aref widths x)))
                  (dotimes(x (- col (current-column writer)))
                    (write-char char (os writer))
                    (incf (slot-value writer 'current-column)))))
              (sepline(char)
                (tab 0 #\space)
                (dotimes(col no-cols)
                  (out "+")
                  (tab (1+ col) char))
                (out "+")
                (write-newline writer)))
        (sepline #\-)
        (dotimes(row no-rows)
          (dotimes(line (aref heights row))
            (tab 0 #\space)
            (do((col 0 (+ col (aref colspans row col))))
               ((>= col no-cols))
              (out "| ")
              (let ((contents (aref contents row col)))
                (when (< line (length contents))
                  (out (elt contents line))))
              (tab (+ col (aref colspans row col))))
            (out "|")
            (write-newline writer))
          (sepline (if (aref header? row) #\= #\-)))))))