;;;; -*- lisp -*-
;;;; $Id: structured-text.lisp,v 1.1 2006/07/18 18:52:49 willijar Exp $
;;;; Copyright (C) 2002 Dr. John A.R. Williams, J.A.R.Williams@blueyonder.co.uk

;;;; This file is part of the Common Lisp Markup Package

;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.

;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.

;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
;;;; or retreive a copy from http://www.gnu.org/licenses/gpl.txt

;;;; An extensible implementation of structured text parsing for Lisp
;;;; Inspired by Structured text from Zope and Python
;;;; Converts structured text to s-expressions for subsequent processing


(in-package :markup)

(defun indentation (string)
  "Find the number of leading spaces. If none, return 0."
  (multiple-value-bind (match left n)
      (scan-str (load-time-value (compile-str "^[:space:]*") t) string)
    (declare (ignore left))
    (if match n 0)))

;; you should see how bad the Python code to do this is.
(defun parse-paragraph-structure(text)
  "Accepts a structured text string and returns a tree structure
   which mimics the structure of the paragraphs as indicated by
   indentation. Structure => (paragraph-text, sub-paragraphs)"
  (let* ((struct '()) ; structure which is to be returned
	 (stack '())) ; stack for tracking current depth
    (dolist (paragraph
	     (regex-split-string
	      (load-time-value
	       (compile-str "(?\\n[:space:]*\\n|\\r\\n[:space:]*\\r\\n)") t)
	      (concatenate 'string
			   #(#\newline #\newline)
			   (expand-tabs text)
			   #(#\newline #\newline))
	      :remove-empty-substr t) struct)
      (let ((indent (indentation paragraph))
	    (paragraph (list paragraph)))
	(while (and stack (<= indent (car (first stack))))
	  (pop stack))
	(if stack
	    (setf (cdr (cdr (first stack)))
		  (append (cdr (cdr (first stack))) (list paragraph)))
	  (setq struct (append struct (list paragraph))))
	(push (cons indent paragraph) stack)))
    struct))

(declaim (inline prepend remove-n))

(defun remove-n(n list)
  "Destructively modify list, removing the nth value from it"
  (let ((p (nthcdr (1- n) list)))
    (rplacd p (cddr p)))
  list)

(defun prepend(value list)
  "Destructively modify list, adding value to the start"
  (rplacd list (cons (car list) (rest list)))
  (rplaca list value)
  list)

;;;; The default set of matching and processing functions
;;;; These are passed the paragraph object, its parent and level
;;;; On input the paragraph is as defined above, a list of its string text
;;;; and the subparagraphs.
;;;; The functions should generally process the subparagraphs by calling
;;;; structured text parse BEFORE processing this paragraph (if appropriate)
;;;; so that when the functions are called they can be sure of the
;;;; structure of the parent object.
;;;; On output they should at least modify
;;;; the list so it starts with an HTML symbol, and the second element
;;;; is text which may have been processed.
;;;; They can assume that the parent object and predecessor in its
;;;; subparagraphs have already been processed.
;;;; the parent structure and predecessor subparagraphs before this element
;;;; may be altered, but not the successors
(defun example-match(paragraph parent)
  "Sub-paragraphs of a paragraph that ends in the word 'example' or the
   word 'examples', or '::' is treated as example code and is output as is."
  (when (and parent (stringp (first parent)))
    (when (scan-str
	   (load-time-value (compile-str "(?::|examples)[:space:]*$") t)
	   (first parent))
      (prepend 'PRE paragraph) )))

(defun list-match(list-symbol paragraph parent &optional preparagraph)
  "If we are a continuation of a previous list of type list-symbol
   we add ourselves to it  else we start a new list of type list-symbol
   and add ourselves to it. In any case we return the new parent object"
  (let ((p (position paragraph parent)))
    (unless
	(when (> p 1)
	  (let ((before (elt parent (1- p))))
	    (when (eq (first before) list-symbol)
	      ;; we are a continuation of a previous list so we move
	      ;; ourselves to it
	      (remove-n p parent)
	      (nconc before (if preparagraph (list preparagraph paragraph)
				(list paragraph))))))
      (setf (elt parent p) (if preparagraph
			       (list list-symbol preparagraph paragraph)
			       (list list-symbol paragraph))))))


(defun ul-match(paragraph parent)
  "A paragraph that begins with a '-', '*', or 'o' is treated as an
unordered list (bullet) element."
  (let ((text (first paragraph)))
    (multiple-value-bind (match left n regs)
	(scan-str (load-time-value
		   (compile-str "^[:space:]*[o\\*-][:space:]+(.+)$") t) text)
      (declare (ignore left n))
      (when match
	(setf (first paragraph) (subseq text (car (elt regs 1))))
	(list-match 'UL paragraph parent)
	(when (cdr paragraph)
	  (setf (first paragraph) (list 'P (first paragraph))))
	(prepend 'LI paragraph) ))))

(defun ol-match(paragraph parent)
  "A paragraph that begins with a sequence of digits followed by a
white-space character is treated as an ordered list element.  A
paragraph that begins with a sequence of sequences, where each
sequence is a sequence of digits or a sequence of letters followed by
a period, is treated as an ordered list element."
  (let ((text (first paragraph)))
    (multiple-value-bind (match left n regs)
	(scan-str
	 (load-time-value
	  (compile-str "^[:space:]*[:digit:]+[.)]+[:space:]+(.+)") t)
	 text)
      (declare (ignore left n))
      (when match
	(setf (first paragraph) (subseq text (car (elt regs 1))))
	(list-match 'OL paragraph parent)
	(when (cdr paragraph)
	  (setf (first paragraph) (list 'P (first paragraph))))
	(prepend 'LI paragraph) ))))

(defun dl-match(paragraph parent)
  " A paragraph with a first line that contains some text, followed
by some white-space and '--' is treated as a descriptive list
element. The leading text is treated as the element title."
  (let ((text (first paragraph)))
    (multiple-value-bind (match left n regs)
	(scan-str
	 (load-time-value
	  (compile-str "^[:space:]*(.*)[:space:]+--(?[:space:]|\\n)+(.*)$") t)
	 text)
      (declare (ignore left n))
      (when match
	(let ((title (subseq text (car (elt regs 1)) (cdr (elt regs 1))))
	      (text (subseq text (car (elt regs 2)))))
	  (list-match 'DL paragraph parent (list 'DT title))
	  (setf (first paragraph) text)
	  (when (cdr paragraph)
	    (setf (first paragraph) (list 'P (first paragraph))))
	  (prepend 'DD paragraph) )))))

(defun table-match(paragraph parent)
  "A paragraph that has blocks of text enclosed in '||' is treated as a
  table. The text blocks correspond to table cells and table rows are
  denoted by newlines. By default the cells are center aligned. A cell
  can span more than one column by preceding a block of text with an
  equivalent number of cell separators '||'. Newlines and '|' cannot
  be a part of the cell text."
  (declare (ignore parent))
  (let ((text (first paragraph)))
    (when (scan-str
	   (load-time-value
	    (compile-str "^[:space:]*\\|\\|.*\\|\\|[:space:]*$") t)
	   text)
      (setf (first paragraph) '(table :border 1 :cellpadding 2))
      (setf (rest paragraph)
	    (mapcar
	     #'(lambda(line)
		 (cons
		  'TR
		  (let ((colspan 0))
		    (mapcan
		     #'(lambda(item)
			 (cond (item
				(let ((result
				       (list (list
					      `(TD :align "center"
						   :colspan ,colspan) item))))
				  (setq colspan 0)
				  result))
			       (t (incf colspan) nil)))
		     (regex-process-string
		      (load-time-value (compile-str "\\|\\|") t)
		      (string-trim " \\r\\n" line)
		      (lambda(s) (declare (ignore s)) nil) )))))
	     (regex-split-string
	      (load-time-value
	       (compile-str "[:space:]*\\r*\\n[:space:]*") t)
	      (string-trim " \\n\\r" text))) )) ))

(defun header-match(paragraph parent)
  "A single-line paragraph whose immediately succeeding paragraphs
are lower level is treated as a title for a new section."
  (declare (ignore parent))
  (when (and (rest paragraph)
	     (not (find #\Linefeed (first paragraph))))
    (rplaca paragraph `(section :title ,(car paragraph)))))

(defun equation-match(paragraph parent)
  (declare (ignore parent))
  (let* ((text (string-trim " \\n\\r" (first paragraph)))
				 (end (1- (length text))))
    (when (and (> end 1) (eq (elt text 0) #\$) (eq (elt text end) #\$))
      (setf (first paragraph) 'EQUATION)
      (setf (rest paragraph) (list (subseq text 1 end)))t )))

(defun default-match(paragraph parent)
  "Default to An ordinary paragraph"
  (declare (ignore parent))
  (prepend 'P  paragraph))

(defparameter *block-hooks*
  (list #'equation-match #'example-match
	#'table-match #'ul-match #'ol-match #'dl-match
        #'header-match #'default-match)
	"A list of functions to be applied to each paragraph in turn.  If
it identifies the paragraph type it should carry out whatever action
and return t. Once a match is found no further tests are made. The
processing functions are given 2 parameters, the paragraph, and the
parent object.")

(defun parse-paragraph-types(paragraphs
			     &key (parent
				   (cons '(div :class "structured-text")
					 paragraphs))
			     (hooks *block-hooks*))
  "parse-paragraph-types accepts a block paragraph structure (from
parse-paragraph-structure) and format Structure =>
(paragraph-text, sub-paragraphs). It formats the subparagraphs and then
applies the list of hooks (default *block-hooks*) until one returns
true. the paragraph list and the contents are destructively modified"
  (dolist (paragraph paragraphs)
    (parse-paragraph-types (rest paragraph)
		 :parent paragraph
		 :hooks hooks)
    (dolist (hook hooks)
      (when (funcall hook paragraph parent)
	    (return))))
   parent)

(defun process-inline-text(things func)
  "Walk through a valid HTML structure, calling func for each inline
text fragment.  The function should a lists of elements will replace
it."
  (if (and (consp things) (not (stringp (car things))))
    (cons
     (first things)
     (mapcan
      #'(lambda (item)
	  (typecase item
	    (string (funcall func item))
	    (list (list (process-inline-text item func)))
	    (t (error "Invalid type ~S in HTML structure" (type-of item)))))
      (rest things)))
  things))

(defun typeface(things)
  "Function deals with the basic typeface operations
     EM,STRONG,U,CODE, which may be nested. "
  (loop
   (let ((no-match t))
     (setq
      things
      (process-inline-text
       things
       #'(lambda(txt)
	   (regex-process-string
	    (load-time-value (compile-str
			      "(?^|[:space:]|[:punct:])+((\\*{1,2}|_|'|\\$)([^:space:].*?(?[:punct:]|[^:space:]))\\2)(?[:space:]*?|$)") t)
	    txt
	    #'(lambda(regs)
		(setq no-match nil)
		(let ((s (elt regs 2)))
		  (list (cond
			  ((equal "*" s) 'EM)
			  ((equal "**" s) 'STRONG)
			  ((equal "_" s) 'U)
			  ((equal "'" s) 'CODE)
			  ((equal "$" s) 'MATH))
			(elt regs 3)))) :match-register 1))))
     (when no-match (return things)))))

(defmacro inline-processor((regex &key (match-register 0)) &body body)
  `(lambda(things)
    (process-inline-text
     things
     #'(lambda(txt)
	 (regex-process-string (load-time-value (compile-str ,regex) t) txt
			       #'(lambda(regs) ,@body)
			       :match-register ,match-register)))))

(defparameter *inline-hooks*
  (list
   (inline-processor
    ("^[:space:]*\.\. (\\[(?([:alnum:]|[_-])+)\\])" )
    (list (list 'A :name (elt regs 2) :id (elt regs 2))
	  (elt regs 1))) ; internal named links
   (inline-processor
    ("(\"([^\"]+?)\"(?:|,[:space:]+)([^:space:]+?)[:punct:]{0,1})([:space:]|$)"
     :match-register 1)
    (list (list 'A :href (elt regs 3)) (elt regs 2)))	;external references
   (inline-processor
    ("[ ,](\\[(?([:alnum:]|[_-])+)\\])(?[:space:]|[:punct:])"
     :match-register 1 )
    (list (list 'A :href (concatenate 'string "#" (elt regs 2)))
	  (elt regs 1)))								; internal references
   #'typeface
   (inline-processor ("([^\\\\])\\\\" :match-register 0 ) (elt regs 1))) ; \\
  "A list of hooks which, given some text, parse it and return a list
   of HTML components")

(defun inline-parse(things &key (hooks *inline-hooks*))
  (dolist (hook hooks)
    (setq things (funcall hook things)))
  things)

(defun parse-structured-text (text &key
				   (paragraph-hooks *block-hooks*)
				   (inline-hooks *inline-hooks*))
  (inline-parse
   (parse-paragraph-types (parse-paragraph-structure text)
			  :hooks paragraph-hooks)
   :hooks inline-hooks))

(defun structured-text-reader(stream c n)
  "Reader macro for structured text.
Without an argument it is parsed as a single line of text
delimited by matching characters e.g.
#!|A single *line* of structured text|
With an argument, the remainder of the line specifies the delimiter
which must be specified on a line by itself and characters are read
line by line"
  (declare (ignore n))
  (let ((term (case c (#\{ #\}) (#\[ #\]) (#\( #\)) (otherwise c))))
    (let ((s-exp
	   (parse-structured-text
	    (with-output-to-string (out)
	      (loop
	       (let ((char (read-char stream)))
		 (cond ((eql char #\\)
			(when (eql (write-char (read-char stream) out) #\\)
			  (write-char #\\ out)))
		       ((eql char #\#)
			(princ (eval (read stream t nil t)) out))
		       ((eql char term) (return))
		       (t (write-char char out)))))))))
      (if (> (length s-exp) 2) s-exp (second s-exp)))))

(defmacro with-structured-text-syntax(&body body)
  `(let ((*readtable* (copy-readtable)))
    (set-dispatch-macro-character #\# #\" #'structured-text-reader *readtable*)
    ,@body))

(defvar *st-doc-text*
  "Structured Text, being a highly Cool Thing, is a tool that Should
Be Known To All. It is used here to allow some formatting of user text
without the user having to know HTML.

Don't worry - it is really quite simple once you know the magical
incantantions. Enjoy!

Structured text is text that uses indentation and simple symbology to
indicate the structure of a document.  A structured string consists of
a sequence of paragraphs separated by one or more blank lines. Each
paragraph has a level which is defined as the minimum indentation of
the paragraph. A paragraph is a sub-paragraph of another paragraph if
the other paragraph is the last preceding paragraph that has a lower
level.

Special symbology is used to indicate special constructs:

 * A single-line paragraph whose immediately succeeding paragraphs are
   lower level is treated as a header.

 * A paragraph that begins with a -, *, or o is treated as an
   unordered list (bullet) element.

 * A paragraph beginning and ending with $ is treated as an equation
   and the content between the $$ will be parsed as a Latex math
   expression, creating an image of the mathematics

 * A paragraph that begins with a sequence of digits followed by a
   white-space character is treated as an ordered list element.

 * A paragraph that begins with a sequence of sequences, where each
   sequence is a sequence of digits or a sequence of letters followed by
   a period, is treated as an ordered list element.

 * A paragraph with a first line that contains some text, followed by some white-space and -- is treated as a descriptive
list element. The leading text is treated as the element title.

 * Sub-paragraphs of a paragraph that ends in the word example or the word examples, or :: is treated as example
code and is output as is.

 * Text enclosed single quotes (with white-space to the left of the first quote and whitespace or puctuation to the right of
the second quote) is treated as example code.

 * Text surrounded by * characters (with white-space to the left of the first * and whitespace or puctuation to the right
of the second *) is emphasized.

 * Text surrounded by ** characters (with white-space to the left of the first ** and whitespace or puctuation to the
right of the second **) is made strong.

 * Text surrounded by _ underscore characters (with whitespace to the left and whitespace or punctuation to the right) is
made underlined.

 * Text in a paragraph surrounded by $ characters will be treated as
   inline mathematics (in Latex markup) and will be rendered as an image

 * Text enclosed by double quotes followed by a colon, a URL, and concluded by punctuation plus white space, or just
white space, is treated as a hyper link. For example:
\"Zope\":http://www.zope.org/ is ...
 Is interpreted as <a href=\"http://www.zope.org/\">Zope</a> is .... Note: This works for relative
as well as absolute URLs.

 * Text enclosed by double quotes followed by a comma, one or more spaces, an absolute URL and concluded by
punctuation plus white space, or just white space, is treated as a hyper link. For example:
\"mail me\", mailto:amos@digicool.com.
Is interpreted as \"<a href=\"mailto:amos@digicool.com\">mail me</a>.\"

 * Text enclosed in brackets which consists only of letters, digits, underscores and dashes is treated as hyper links within
the document. For example:
As demonstrated by Smith [a12] this technique is quite effective. [r1]
Is interpreted as ... by Smith <a href=\"#a12\">[12]</a> this .... Together with the next rule this
allows easy coding of references or end notes.

 * Text enclosed in brackets which is preceded by the start of a line, two periods and a space is treated as a named link.
For example:
.. [a12] \"Effective Techniques\" Smith, Joe
Is interpreted as <a name=\"a12\">[12]</a> \"Effective Techniques\" .... [r2] Together with the
previous rule this allows easy coding of references or end notes.

.. [r1] According to the HTML 4.0 specification identifiers must start with a letter so using references like [12] is a bit of a
no-no, though it does work (at least in all the browsers I've tried). The StructuredText Python module should probably be
updated to do something nice like prepend ref whenver it sees only numbers in a link.

.. [r2] The use of name here is deprecated, id should be used instead. Not there is much you as the document author can do
about this since it's StructuredText that is generating the HTML (and it probably should be modified to do the right thing) but
at least when somebody comes to nitpick you can say \"Yes, I know. Everybody knows about this, now go away and let me get
some work done before I have to show just how strong my Kung Fu really is.\""
  "The user documentation as Structured Text!")
